//
// File:        MblkLinAdv.C
// Package:     SAMRAI application
// Copyright:   (c) 1997-2008 Lawrence Livermore National Security, LLC
// Revision:    $LastChangedRevision: 2147 $
// Modified:    $LastChangedDate: 2008-04-23 16:48:12 -0700 (Wed, 23 Apr 2008) $
// Description: Numerical routines for single patch in linear advection ex.
//
#include "MblkLinAdv.h"

#include <iostream>
#include <iomanip>
#include <fstream>

#ifndef LACKS_SSTREAM
#ifndef included_sstream
#define included_sstream
#include <sstream>
#endif
#else
#ifndef included_strstream
#define included_strstream
#include <strstream.h>
#endif
#endif

using namespace std;

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>

#include "tbox/ArenaManager.h"
#include "tbox/Array.h"
#include "BoundaryBox.h"
#include "BoxArray.h"
#include "BoundaryBox.h"
#include "CellData.h"
#include "CellIndex.h"
#include "CellIterator.h"
#include "CellVariable.h"
#include "CellDoubleLinearTimeInterpolateOp.h"
#include "CoarsenOperator.h"
#include "NodeIndex.h"
#include "SideData.h"
#include "SideIndex.h"
#include "SideVariable.h"
#include "BlockGridGeometry.h"
#include "BlockPatchGeometry.h"
#include "tbox/InputManager.h"
#include "Index.h"
#include "LoadBalancer.h"
#include "NodeData.h"
#include "NodeIndex.h"
#include "NodeIterator.h"
#include "NodeDoubleInjection.h"
#include "NodeDoubleLinearTimeInterpolateOp.h"
#include "RefineOperator.h"
#include "BlockPatchGeometry.h"
#include "tbox/PIO.h"
#include "tbox/RestartManager.h"
#include "tbox/Utilities.h"
#include "tbox/MathUtilities.h"
#include "TimeInterpolateOperator.h"
#include "VariableDatabase.h"

//integer constants for boundary conditions
#define CHECK_BDRY_DATA  (0)
#include "CartesianBoundaryDefines.h"

//integer constant for debugging improperly set boundary dat
#define BOGUS_BDRY_DATA   (-9999)

// routines for managing boundary data
#if (NDIM == 2)
#include "SkeletonBoundaryUtilities2.h"
#endif
#if (NDIM == 3)
#include "SkeletonBoundaryUtilities3.h"
#endif

// Depth of the advected variable
#define DEPTH           (1)

// Number of ghosts cells used for each variable quantity
#define CELLG           (4)
#define FACEG           (4)
#define FLUXG           (0)
#define NODEG           (0)

// defines for initialization
#define PIECEWISE_CONSTANT_X        (10)
#define PIECEWISE_CONSTANT_Y        (11)
#define PIECEWISE_CONSTANT_Z        (12)
#define SINE_CONSTANT_X             (20)
#define SINE_CONSTANT_Y             (21)
#define SINE_CONSTANT_Z             (22)
#define SPHERE                      (40)

// defines for Riemann solver used in Godunov flux calculation
#define APPROX_RIEM_SOLVE   (20) // Colella-Glaz approx Riemann solver
#define EXACT_RIEM_SOLVE    (21) // Exact Riemann solver
#define HLLC_RIEM_SOLVE     (22) // Harten, Lax, van Leer approx Riemann solver

// defines for cell tagging routines
#define RICHARDSON_NEWLY_TAGGED (-10)
#define RICHARDSON_ALREADY_TAGGED (-11)
#ifndef TRUE
#define TRUE (1)
#endif
#ifndef FALSE
#define FALSE (0)
#endif

// Version of MblkLinAdv restart file data
#define MBLKLINADV_VERSION (3)

//
// some extra defines for C code
//
#define real8 double
#define POLY3(i,j,k,imin,jmin,kmin,nx,nxny) ( (i-imin) + (j-jmin)*(nx) + (k-kmin)*(nxny) )
#define MAX( a, b ) ( a > b ? a : b )
#define MIN( a, b ) ( a < b ? a : b )

/*
*************************************************************
*
* Extra non-class inlined C functions used by class methods
*
*************************************************************
*/
//
// calculate the flux through a face
//
inline
real8 UpwindFlux( const real8 x1, const real8 x2,
		  const real8 x3, const real8 x4,
		  const real8 y1, const real8 y2,
		  const real8 y3, const real8 y4,
		  const real8 z1, const real8 z2,
		  const real8 z3, const real8 z4,
		  real8 u, real8 v, real8 w,
		  real8 psiLo, real8 psiHi )
{
   real8 dx31 = x3 - x1;
   real8 dx42 = x4 - x2;

   real8 dy31 = y3 - y1;
   real8 dy42 = y4 - y2;

   real8 dz31 = z3 - z1;
   real8 dz42 = z4 - z2;

   real8 Ax = 0.5 * (dy42 * dz31 - dz42 * dy31);
   real8 Ay = 0.5 * (dz42 * dx31 - dx42 * dz31);
   real8 Az = 0.5 * (dx42 * dy31 - dy42 * dx31);

   real8 Audotn = Ax*u + Ay*v + Az*w;

   real8 flux = ( Audotn > 0.0 ? psiLo : psiHi )*Audotn;

   return flux;
}


//
// calculate the flux through a face, assuming a velocity in the radial direction
//
inline
real8 UpwindFluxRadial( const real8 x1, const real8 x2,
			const real8 x3, const real8 x4,
			const real8 y1, const real8 y2,
			const real8 y3, const real8 y4,
			const real8 z1, const real8 z2,
			const real8 z3, const real8 z4,
			real8 u0,
			real8 psiLo, real8 psiHi )
{
   // --------- set the velocity
   real8 xm = 0.25*(x1 + x2 + x3 + x4);
   real8 ym = 0.25*(y1 + y2 + y3 + y4);
   real8 zm = 0.25*(z1 + z2 + z3 + z4);
   real8 xnorm = sqrt( xm*xm + ym*ym + zm*zm );

   real8 u = u0 * xm / xnorm;
   real8 v = u0 * ym / xnorm;
   real8 w = u0 * zm / xnorm;

   // --------- set the flux
   real8 dx31 = x3 - x1;
   real8 dx42 = x4 - x2;

   real8 dy31 = y3 - y1;
   real8 dy42 = y4 - y2;

   real8 dz31 = z3 - z1;
   real8 dz42 = z4 - z2;

   real8 Ax = 0.5 * (dy42 * dz31 - dz42 * dy31);
   real8 Ay = 0.5 * (dz42 * dx31 - dx42 * dz31);
   real8 Az = 0.5 * (dx42 * dy31 - dy42 * dx31);

   real8 Audotn = Ax*u + Ay*v + Az*w;

   real8 flux = ( Audotn > 0.0 ? psiLo : psiHi )*Audotn;

   return flux;
}


//
// calculate the volume of a hexahedral element
//
inline
real8 UpwindVolume(const real8 x0, const real8 x1, const real8 x2, const real8 x3,
		   const real8 x4, const real8 x5, const real8 x6, const real8 x7,
		   const real8 y0, const real8 y1, const real8 y2, const real8 y3,
		   const real8 y4, const real8 y5, const real8 y6, const real8 y7,
		   const real8 z0, const real8 z1, const real8 z2, const real8 z3,
		   const real8 z4, const real8 z5, const real8 z6, const real8 z7)
{
   const real8 twelfth = 1.0/12.0;
   real8 volume, s1234, s5678, s1265, s4378, s2376, s1485;

   s1234 =
      (x1+x2)*( (y0+y1)*(z2+z3) - (z0+z1)*(y2+y3) ) +
      (y1+y2)*( (z0+z1)*(x2+x3) - (x0+x1)*(z2+z3) ) +
      (z1+z2)*( (x0+x1)*(y2+y3) - (y0+y1)*(x2+x3) );

   s5678 =
      (x5+x6)*( (y4+y5)*(z6+z7) - (z4+z5)*(y6+y7) ) +
      (y5+y6)*( (z4+z5)*(x6+x7) - (x4+x5)*(z6+z7) ) +
      (z5+z6)*( (x4+x5)*(y6+y7) - (y4+y5)*(x6+x7) );

   s1265 =
      (x1+x5)*( (y0+y1)*(z5+z4) - (z0+z1)*(y5+y4) ) +
      (y1+y5)*( (z0+z1)*(x5+x4) - (x0+x1)*(z5+z4) ) +
      (z1+z5)*( (x0+x1)*(y5+y4) - (y0+y1)*(x5+x4) );

   s4378 =
      (x2+x6)*( (y3+y2)*(z6+z7) - (z3+z2)*(y6+y7) ) +
      (y2+y6)*( (z3+z2)*(x6+x7) - (x3+x2)*(z6+z7) ) +
      (z2+z6)*( (x3+x2)*(y6+y7) - (y3+y2)*(x6+x7) );

   s2376 =
      (x2+x6)*( (y1+y2)*(z6+z5) - (z1+z2)*(y6+y5) ) +
      (y2+y6)*( (z1+z2)*(x6+x5) - (x1+x2)*(z6+z5) ) +
      (z2+z6)*( (x1+x2)*(y6+y5) - (y1+y2)*(x6+x5) );

   s1485 =
      (x3+x7)*( (y0+y3)*(z7+z4) - (z0+z3)*(y7+y4) ) +
      (y3+y7)*( (z0+z3)*(x7+x4) - (x0+x3)*(z7+z4) ) +
      (z3+z7)*( (x0+x3)*(y7+y4) - (y0+y3)*(x7+x4) );

   volume =  (s1234 - s5678 - s1265 + s4378 - s2376 + s1485) * twelfth ;
   return volume ;
}

//
// compute the area of a face
//
inline real8
UpwindAreaFace( const real8 x0, const real8 x1,
		const real8 x2, const real8 x3,
		const real8 y0, const real8 y1,
		const real8 y2, const real8 y3,
		const real8 z0, const real8 z1,
		const real8 z2, const real8 z3)
{
   real8 fx = (x2 - x0) - (x3 - x1);
   real8 fy = (y2 - y0) - (y3 - y1);
   real8 fz = (z2 - z0) - (z3 - z1);
   real8 gx = (x2 - x0) + (x3 - x1);
   real8 gy = (y2 - y0) + (y3 - y1);
   real8 gz = (z2 - z0) + (z3 - z1);
   real8 area =
      (fx * fx + fy * fy + fz * fz) *
      (gx * gx + gy * gy + gz * gz) -
      (fx * gx + fy * gy + fz * gz) *
      (fx * gx + fy * gy + fz * gz);
   return area ;
}

//
// compute a characteristic length
//
inline real8
UpwindCharacteristicLength( const real8 x[8], const real8 y[8],
			    const real8 z[8], const real8 volume)
{
   real8 a, charLength = 0.0;

   a = UpwindAreaFace(x[0],x[1],x[2],x[3],
		      y[0],y[1],y[2],y[3],
		      z[0],z[1],z[2],z[3]) ;
   charLength = MAX(a,charLength) ;

   a = UpwindAreaFace(x[4],x[5],x[6],x[7],
		      y[4],y[5],y[6],y[7],
		      z[4],z[5],z[6],z[7]) ;
   charLength = MAX(a,charLength) ;

   a = UpwindAreaFace(x[0],x[1],x[5],x[4],
		      y[0],y[1],y[5],y[4],
		      z[0],z[1],z[5],z[4]) ;
   charLength = MAX(a,charLength) ;

   a = UpwindAreaFace(x[1],x[2],x[6],x[5],
		      y[1],y[2],y[6],y[5],
		      z[1],z[2],z[6],z[5]) ;
   charLength = MAX(a,charLength) ;

   a = UpwindAreaFace(x[2],x[3],x[7],x[6],
		      y[2],y[3],y[7],y[6],
		      z[2],z[3],z[7],z[6]) ;
   charLength = MAX(a,charLength) ;

   a = UpwindAreaFace(x[3],x[0],x[4],x[7],
		      y[3],y[0],y[4],y[7],
		      z[3],z[0],z[4],z[7]) ;
   charLength = MAX(a,charLength) ;

   charLength = 4.0 * volume / sqrt(charLength);

   return charLength;
}


///
/// the cartesian uniform grid monotonic slope finder
///
inline void my_slopesCart( double psi, double pim, double pip, double pjm,
                           double pjp, double pkm, double pkp,
                           double &pxi, double &peta, double &pzeta )
{
   real8 del, sfp, sbm, scale;
   real8 elDenp, elDenC, elDenm;
   real8 sfact = 0.25;  // due to the fact that xi ranges from -1 to 1
   elDenC = psi;

   //
   // xi
   //
   elDenp = pip;
   elDenm = pim;

   del = sfact*(elDenp - elDenm) + 1.e-80;
   sfp = (elDenp -elDenC ) * 2. / del;
   sbm = (elDenC -elDenm ) * 2. / del;

   scale = MIN(sfp,sbm) ;
   scale = ( scale   > 1.0   ? 1.0 : scale );
   scale = ( scale   < 0.0   ? 0.0 : scale );
   scale = ( sfp*sbm < 0.0   ? 0.0 : scale );

   pxi = del*scale;  // xi, eta, zeta vary from -1 to 1

   //
   // eta
   //
   elDenp = pjp;
   elDenm = pjm;

   del = sfact*(elDenp - elDenm) + 1.e-80;
   sfp = (elDenp -elDenC ) * 2. / del;
   sbm = (elDenC -elDenm ) * 2. / del;

   scale = MIN(sfp,sbm) ;
   scale = ( scale   > 1.0   ? 1.0 : scale );
   scale = ( scale   < 0.0   ? 0.0 : scale );
   scale = ( sfp*sbm < 0.0   ? 0.0 : scale );

   peta = del*scale;

   //
   // eta
   //
   elDenp = pkp;
   elDenm = pkm;

   del = sfact*(elDenp - elDenm) + 1.e-80;
   sfp = (elDenp -elDenC ) * 2. / del;
   sbm = (elDenC -elDenm ) * 2. / del;

   scale = MIN(sfp,sbm) ;
   scale = ( scale   > 1.0   ? 1.0 : scale );
   scale = ( scale   < 0.0   ? 0.0 : scale );
   scale = ( sfp*sbm < 0.0   ? 0.0 : scale );

   pzeta = del*scale;
}





///
/// the non-uniform grid monotonic slope finder
///
inline void my_slopes( double psi, double pim, double pip, double pjm,
                       double pjp, double pkm, double pkp,
                       double w_i, double w_ip, double w_im, double w_jp,
                       double w_jm, double w_kp, double w_km,
                       double &pxi, double &peta, double &pzeta )
{
   real8 sumf, sumb;
   real8 elDenm, elDenp, del, sfp, sbm, scale;

   real8 elDenC = psi;
   real8 scale_fact = 1.0;
   real8 slope_fact = 1.0;

   //
   // compute weight functions
   //
   real8 volzrc = w_i;

   real8 volzrxim   = w_im;
   real8 volzrxip   = w_ip;
   real8 volzretam  = w_jm;
   real8 volzretap  = w_jp;
   real8 volzrzetam = w_km;
   real8 volzrzetap = w_kp;

   sumf = volzrc + volzrxip;
   sumb = volzrc + volzrxim;
   real8 wgtxi1 = volzrc / sumf;
   real8 wgtxi2 = volzrxip / sumf;
   real8 wgtxi3 = volzrc / sumb;
   real8 wgtxi4 = volzrxim / sumb;

   sumf = volzrc + volzretap;
   sumb = volzrc + volzretam;
   real8 wgteta1 = volzrc / sumf;
   real8 wgteta2 = volzretap / sumf;
   real8 wgteta3 = volzrc / sumb;
   real8 wgteta4 = volzretam / sumb;

   sumf = volzrc + volzrzetap;
   sumb = volzrc + volzrzetam;
   real8 wgtzeta1 = volzrc / sumf;
   real8 wgtzeta2 = volzrzetap / sumf;
   real8 wgtzeta3 = volzrc / sumb;
   real8 wgtzeta4 = volzrzetam / sumb;


   elDenm = pim;
   elDenp = pip;

   del = ( wgtxi2 * elDenp + wgtxi1 * elDenC
           - wgtxi4 * elDenm - wgtxi3 * elDenC) + 1e-80;

   sfp = (elDenp - elDenC ) * scale_fact / del;
   sbm = (elDenC - elDenm ) * scale_fact / del;

   scale = MIN(sfp,sbm) ;
   if (scale > 1.)
      scale = 1.0;
   else if (scale < 0.)
      scale = 0.;

   if ((sfp*sbm)<0.0) scale = 0.;

   pxi = slope_fact * del * scale;

   // --------------------------------- eta

   elDenm = pjm;
   elDenp = pjp;

   del = (wgteta2 * elDenp + wgteta1 * elDenC
          - wgteta4 * elDenm - wgteta3 * elDenC) + 1e-80;

   sfp = (elDenp - elDenC ) * scale_fact / del;
   sbm = (elDenC - elDenm ) * scale_fact / del;

   scale = MIN(sfp,sbm) ;
   if (scale > 1.)
      scale = 1.0;
   else if (scale < 0.)
      scale = 0.;

   if ((sfp*sbm)<0.0) scale = 0.;

   peta  = slope_fact * del * scale;

   // --------------------------------- zeta

   elDenm = pkm;
   elDenp = pkp;

   del = (wgtzeta2 * elDenp + wgtzeta1 * elDenC
          - wgtzeta4 * elDenm - wgtzeta3 * elDenC) + 1e-80;

   sfp = (elDenp - elDenC ) * scale_fact / del;
   sbm = (elDenC - elDenm ) * scale_fact / del;

   scale = MIN(sfp,sbm) ;
   if (scale > 1.)
      scale = 1.0;
   else if (scale < 0.)
      scale = 0.;

   if ((sfp*sbm)<0.0) scale = 0.;

   pzeta = slope_fact * del * scale;
}



/*
*************************************************************************
*                                                                       *
* The constructor for MblkLinAdv class sets data members to defualt values, *
* creates variables that define the solution state for the linear       *
* advection equation.
*                                                                       *
* After default values are set, this routine calls getFromRestart()     *
* if execution from a restart file is specified.  Finally,              *
* getFromInput() is called to read values from the given input          *
* database (potentially overriding those found in the restart file).    *
*                                                                       *
*************************************************************************
*/

MblkLinAdv::MblkLinAdv(
   const string& object_name,
   tbox::Pointer<tbox::Database> input_db,
   tbox::Array< tbox::Pointer<hier::GridGeometry<NDIM> > >& grid_geoms)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   TBOX_ASSERT(!object_name.empty());
   TBOX_ASSERT(!input_db.isNull());
   TBOX_ASSERT(grid_geoms.getSize() > 0);
#endif

   d_object_name = object_name;
   tbox::RestartManager::getManager()->registerRestartItem(d_object_name, this);

   d_grid_geometries = grid_geoms;

   /*
    * Setup MblkGeometry object to manage construction of mapped grids
    */
   d_mblk_geometry = new MblkGeometry("MblkGeometry",
                                      input_db,
                                      grid_geoms.getSize());

   d_use_nonuniform_workload = false;

   /*
    * hier::Variable quantities that define state of linear advection problem.
    */
   d_uval         = new pdat::CellVariable<NDIM,double>("uval",DEPTH);
   d_vol          = new pdat::CellVariable<NDIM,double>("vol",1);
   d_flux         = new pdat::SideVariable<NDIM,double>("flux",1);
   d_xyz          = new pdat::NodeVariable<NDIM,double>("xyz",NDIM);

   d_dx_set = false;

   /*
    * Default parameters for the numerical method.
    */
   d_godunov_order = 1;
   d_corner_transport = "CORNER_TRANSPORT_1";
#ifdef DEBUG_CHECK_ASSERTIONS
   TBOX_ASSERT(CELLG == FACEG);
#endif
   d_nghosts = hier::IntVector<NDIM>(CELLG);
   d_fluxghosts = hier::IntVector<NDIM>(FLUXG);
   d_nodeghosts = hier::IntVector<NDIM>(NODEG);

   /*
    * Defaults for problem type and initial data.
    */
   d_data_problem_int = tbox::MathUtilities<int>::getMax();

   int k;

   // SPHERE problem...
   d_radius = tbox::MathUtilities<double>::getSignalingNaN();
   tbox::MathUtilities<double>::setArrayToSignalingNaN(d_center, NDIM);
   d_uval_inside = tbox::MathUtilities<double>::getSignalingNaN();
   d_uval_outside = tbox::MathUtilities<double>::getSignalingNaN();

   d_number_of_intervals = 0;
   d_front_position.resizeArray(0);
   d_interval_uval.resizeArray(0);

   // SINE problem
   d_amplitude = 0.;
   for (k = 0; k < NDIM; k++) d_frequency[k] = 0.;

   /*
    * Defaults for boundary conditions. Set to bogus values
    * for error checking.
    */

#if (NDIM == 2)
   d_scalar_bdry_edge_conds.resizeArray(NUM_2D_EDGES);
   for (int ei = 0; ei < NUM_2D_EDGES; ei++) {
      d_scalar_bdry_edge_conds[ei] = BOGUS_BDRY_DATA;
   }

   d_scalar_bdry_node_conds.resizeArray(NUM_2D_NODES);
   d_node_bdry_edge.resizeArray(NUM_2D_NODES);

   for (int ni = 0; ni < NUM_2D_NODES; ni++) {
      d_scalar_bdry_node_conds[ni] = BOGUS_BDRY_DATA;
      d_node_bdry_edge[ni] = BOGUS_BDRY_DATA;
   }

   d_bdry_edge_uval.resizeArray(NUM_2D_EDGES);
   tbox::MathUtilities<double>::setArrayToSignalingNaN(d_bdry_edge_uval);
#endif
#if (NDIM == 3)
   d_scalar_bdry_face_conds.resizeArray(NUM_3D_FACES);
   for (int fi = 0; fi < NUM_3D_FACES; fi++) {
      d_scalar_bdry_face_conds[fi] = BOGUS_BDRY_DATA;
   }

   d_scalar_bdry_edge_conds.resizeArray(NUM_3D_EDGES);
   d_edge_bdry_face.resizeArray(NUM_3D_EDGES);
   for (int ei = 0; ei < NUM_3D_EDGES; ei++) {
      d_scalar_bdry_edge_conds[ei] = BOGUS_BDRY_DATA;
      d_edge_bdry_face[ei] = BOGUS_BDRY_DATA;
   }

   d_scalar_bdry_node_conds.resizeArray(NUM_3D_NODES);
   d_node_bdry_face.resizeArray(NUM_3D_NODES);

   for (int ni = 0; ni < NUM_3D_NODES; ni++) {
      d_scalar_bdry_node_conds[ni] = BOGUS_BDRY_DATA;
      d_node_bdry_face[ni] = BOGUS_BDRY_DATA;
   }

   d_bdry_face_uval.resizeArray(NUM_3D_FACES);
   tbox::MathUtilities<double>::setArrayToSignalingNaN(d_bdry_face_uval);
#endif

   /*
    * Initialize object with data read from given input/restart databases.
    */
   bool is_from_restart = tbox::RestartManager::getManager()->isFromRestart();
   if (is_from_restart){
      getFromRestart();
   }
   getFromInput(input_db, is_from_restart);

   /*
    * Set problem data to values read from input/restart.
    */

   if (d_data_problem == "PIECEWISE_CONSTANT_X") {
      d_data_problem_int = PIECEWISE_CONSTANT_X;
   } else if (d_data_problem == "PIECEWISE_CONSTANT_Y") {
      d_data_problem_int = PIECEWISE_CONSTANT_Y;
   } else if (d_data_problem == "PIECEWISE_CONSTANT_Z") {
      d_data_problem_int = PIECEWISE_CONSTANT_Z;
   } else if (d_data_problem == "SINE_CONSTANT_X") {
      d_data_problem_int = SINE_CONSTANT_X;
   } else if (d_data_problem == "SINE_CONSTANT_Y") {
      d_data_problem_int = SINE_CONSTANT_Y;
   } else if (d_data_problem == "SINE_CONSTANT_Z") {
      d_data_problem_int = SINE_CONSTANT_Z;
   } else if (d_data_problem == "SPHERE") {
      d_data_problem_int = SPHERE;
   } else {
      TBOX_ERROR(d_object_name << ": "
         << "Unknown d_data_problem string = "
         << d_data_problem << " encountered in constructor" << endl);
   }

   /*
    * Postprocess boundary data from input/restart values.  Note: scalar
    * quantity in this problem cannot have reflective boundary conditions
    * so we reset them to FLOW.
    */
#if (NDIM == 2)
   for (int i = 0; i < NUM_2D_EDGES; i++) {
      if (d_scalar_bdry_edge_conds[i] == REFLECT_BC) {
         d_scalar_bdry_edge_conds[i] = FLOW_BC;
      }
   }

   for (int i = 0; i < NUM_2D_NODES; i++) {
      if (d_scalar_bdry_node_conds[i] == XREFLECT_BC) {
         d_scalar_bdry_node_conds[i] = XFLOW_BC;
      }
      if (d_scalar_bdry_node_conds[i] == YREFLECT_BC) {
         d_scalar_bdry_node_conds[i] = YFLOW_BC;
      }

      if (d_scalar_bdry_node_conds[i] != BOGUS_BDRY_DATA) {
         d_node_bdry_edge[i] =
	    SkeletonBoundaryUtilities2::getEdgeLocationForNodeBdry(
                                            i, d_scalar_bdry_node_conds[i]);
      }
   }
#endif
#if (NDIM == 3)
   for (int i = 0; i < NUM_3D_FACES; i++) {
      if (d_scalar_bdry_face_conds[i] == REFLECT_BC) {
         d_scalar_bdry_face_conds[i] = FLOW_BC;
      }
   }

   for (int i = 0; i < NUM_3D_EDGES; i++) {
      if (d_scalar_bdry_edge_conds[i] == XREFLECT_BC) {
         d_scalar_bdry_edge_conds[i] = XFLOW_BC;
      }
      if (d_scalar_bdry_edge_conds[i] == YREFLECT_BC) {
         d_scalar_bdry_edge_conds[i] = YFLOW_BC;
      }
      if (d_scalar_bdry_edge_conds[i] == ZREFLECT_BC) {
         d_scalar_bdry_edge_conds[i] = ZFLOW_BC;
      }

      if (d_scalar_bdry_edge_conds[i] != BOGUS_BDRY_DATA) {
         d_edge_bdry_face[i] =
            SkeletonBoundaryUtilities3::getFaceLocationForEdgeBdry(
                                            i, d_scalar_bdry_edge_conds[i]);
      }
   }

   for (int i = 0; i < NUM_3D_NODES; i++) {
      if (d_scalar_bdry_node_conds[i] == XREFLECT_BC) {
         d_scalar_bdry_node_conds[i] = XFLOW_BC;
      }
      if (d_scalar_bdry_node_conds[i] == YREFLECT_BC) {
         d_scalar_bdry_node_conds[i] = YFLOW_BC;
      }
      if (d_scalar_bdry_node_conds[i] == ZREFLECT_BC) {
         d_scalar_bdry_node_conds[i] = ZFLOW_BC;
      }

      if (d_scalar_bdry_node_conds[i] != BOGUS_BDRY_DATA) {
         d_node_bdry_face[i] =
            SkeletonBoundaryUtilities3::getFaceLocationForNodeBdry(
                                            i, d_scalar_bdry_node_conds[i]);
      }
   }

#endif

}

/*
*************************************************************************
*                                                                       *
* Empty destructor for MblkLinAdv class.                                    *
*                                                                       *
*************************************************************************
*/

MblkLinAdv::~MblkLinAdv() {
   if (d_mblk_geometry) delete d_mblk_geometry;
}

/*
*************************************************************************
*                                                                       *
* Register conserved variable (u) (i.e., solution state variable) and   *
* flux variable with hyperbolic integrator that manages storage for     *
* those quantities.  Also, register plot data with Vizamrai or VisIt.   *
*                                                                       *
*************************************************************************
*/

void MblkLinAdv::registerModelVariables(
   MblkHyperbolicLevelIntegrator* integrator)
{

#ifdef DEBUG_CHECK_ASSERTIONS
   TBOX_ASSERT(integrator != (MblkHyperbolicLevelIntegrator*)NULL);
   TBOX_ASSERT(CELLG == FACEG);
#endif

   d_cell_cons_linear_refine_op =
      new SkeletonCellDoubleConservativeLinearRefine();
   d_cell_cons_coarsen_op =
      new SkeletonCellDoubleWeightedAverage();
   d_cell_time_interp_op =
      new pdat::CellDoubleLinearTimeInterpolateOp<NDIM>();
   d_side_cons_coarsen_op =
      new SkeletonOutersideDoubleWeightedAverage();

   // Note that the Node linear refine operator is null for this case,
   // which is OK because the only node data is the grid coordinates,
   // which we explicitly set on any new patch
   tbox::Pointer<xfer::RefineOperator<NDIM> > node_linear_refine_op;
   tbox::Pointer<pdat::NodeDoubleInjection<NDIM> > node_cons_coarsen_op =
      new pdat::NodeDoubleInjection<NDIM>();

   integrator->registerVariable(d_uval, d_nghosts,
                                MblkHyperbolicLevelIntegrator::TIME_DEP,
                                d_cell_cons_coarsen_op,
                                d_cell_cons_linear_refine_op,
                                d_cell_time_interp_op);

   integrator->registerVariable(d_vol, d_nghosts,
                                MblkHyperbolicLevelIntegrator::TIME_DEP,
                                d_cell_cons_coarsen_op,
                                d_cell_cons_linear_refine_op,
                                d_cell_time_interp_op);

   integrator->registerVariable(d_flux, d_fluxghosts,
				MblkHyperbolicLevelIntegrator::FLUX,
                                d_side_cons_coarsen_op);

#if 1
   tbox::Pointer<xfer::TimeInterpolateOperator<NDIM> > node_time_interp_op =
      new pdat::NodeDoubleLinearTimeInterpolateOp<NDIM>();
   integrator->registerVariable(d_xyz, d_nodeghosts,
                                MblkHyperbolicLevelIntegrator::TIME_DEP,
                                node_cons_coarsen_op,
                                node_linear_refine_op,
                                node_time_interp_op);
#endif
#if 0
   integrator->registerVariable(d_xyz, d_nodeghosts,
                                MblkHyperbolicLevelIntegrator::INPUT,
                                node_cons_coarsen_op,
                                node_linear_refine_op);
#endif


   hier::VariableDatabase<NDIM>* vardb = hier::VariableDatabase<NDIM>::getDatabase();



#ifdef HAVE_HDF5
   if (!(d_visit_writer.isNull())) {
      d_visit_writer->
         registerPlotQuantity("U",
                              "SCALAR",
                              vardb->mapVariableAndContextToIndex(
                                 d_uval, integrator->getPlotContext()));
      d_visit_writer->
         registerPlotQuantity("vol",
                              "SCALAR",
                              vardb->mapVariableAndContextToIndex(
                                 d_vol, integrator->getPlotContext()));
      d_visit_writer->
         registerNodeCoordinates(
            vardb->mapVariableAndContextToIndex(
               d_xyz, integrator->getPlotContext()));
   }

   if (d_visit_writer.isNull()) {
      TBOX_WARNING(d_object_name << ": registerModelVariables()"
                   << "\nVisit data writer was"
                   << "\nregistered.  Consequently, no plot data will"
                   << "\nbe written." << endl);
   }
#endif

}


/*
*************************************************************************
*                                                                       *
* Set initial data for solution variables on patch interior.            *
* This routine is called whenever a new patch is introduced to the      *
* AMR patch hierarchy.  Note that the routine does nothing unless       *
* we are at the initial time.  In all other cases, conservative         *
* interpolation from coarser levels and copies from patches at the      *
* same mesh resolution are sufficient to set data.                      *
*                                                                       *
*************************************************************************
*/
void MblkLinAdv::initializeDataOnPatch(hier::Patch<NDIM>& patch,
                                   const double data_time,
                                   const bool initial_time)
{
   //(void) data_time;

   /*
    * Build the mapped grid on the patch.
    */
   int block_number = getBlockNumber();
   int level_number =  patch.getPatchLevelNumber();
   setMappedGridOnPatch(patch, level_number, block_number);

   /*
    * Set the dx in the operators
    */
   double dx[NDIM];
   d_mblk_geometry->getDx(level_number,dx);
   d_dx_set = true;

   d_cell_cons_linear_refine_op->setDx(level_number,dx);
   d_cell_cons_coarsen_op->setDx(level_number,dx);
   d_side_cons_coarsen_op->setDx(level_number,dx);

   if (initial_time) {

      tbox::Pointer< pdat::CellData<NDIM,double> > uval =
         patch.getPatchData(d_uval, getDataContext());
      tbox::Pointer< pdat::CellData<NDIM,double> > vol =
         patch.getPatchData(d_vol, getDataContext());
      tbox::Pointer< pdat::NodeData<NDIM,double> > xyz =
         patch.getPatchData(d_xyz, getDataContext());

#ifdef DEBUG_CHECK_ASSERTIONS
      TBOX_ASSERT(!uval.isNull());
      TBOX_ASSERT(!vol.isNull());
      TBOX_ASSERT(!xyz.isNull());
      TBOX_ASSERT(uval->getGhostCellWidth() == vol->getGhostCellWidth());
#endif

      hier::IntVector<NDIM> uval_ghosts = uval->getGhostCellWidth();
      hier::IntVector<NDIM> xyz_ghosts = xyz->getGhostCellWidth();

      const hier::Index<NDIM> ifirst=patch.getBox().lower();
      const hier::Index<NDIM> ilast =patch.getBox().upper();

      int imin = ifirst(0) - uval_ghosts(0);
      int imax = ilast(0)  + uval_ghosts(0);
      int jmin = ifirst(1) - uval_ghosts(1);
      int jmax = ilast(1)  + uval_ghosts(1);
      int kmin = ifirst(2) - uval_ghosts(2);
      //int kmax = ilast(2)  + uval_ghosts(2);
      int nx   = imax - imin + 1;
      int ny   = jmax - jmin + 1;
      // int nz   = kmax - kmin + 1;
      int nxny = nx*ny;
      // int nel  = nx*ny*nz;

      int nd_imin = ifirst(0)     - xyz_ghosts(0);
      int nd_imax = ilast(0)  + 1 + xyz_ghosts(0);
      int nd_jmin = ifirst(1)     - xyz_ghosts(1);
      int nd_jmax = ilast(1)  + 1 + xyz_ghosts(1);
      int nd_kmin = ifirst(2)     - xyz_ghosts(2);
      //int nd_kmax = ilast(2)  + 1 + xyz_ghosts(2);
      int nd_nx   = nd_imax - nd_imin + 1;
      int nd_ny   = nd_jmax - nd_jmin + 1;
      // int nd_nz   = nd_kmax - nd_kmin + 1;
      int nd_nxny = nd_nx*nd_ny;
      // int nd_nel  = nd_nx*nd_ny*nd_nz;

      //
      // get the pointers
      //
      double *psi = uval->getPointer();
      double *cvol = vol->getPointer();

      double *x = xyz->getPointer(0);
      double *y = xyz->getPointer(1);
      double *z = xyz->getPointer(2);

      //
      // compute the source due to the upwind method
      //
      for ( int k = ifirst(2); k <= ilast(2); k++ ) {
	 for ( int j = ifirst(1); j <= ilast(1); j++ ) {
	    for ( int i = ifirst(0); i <= ilast(0); i++ ) {

               int cind = POLY3(i,j,k,imin,jmin,kmin,nx,nxny);

	       int n1 = POLY3(i,j,k,nd_imin,nd_jmin,nd_kmin,nd_nx,nd_nxny);
	       int n2 = n1 + 1;
	       int n3 = n1 + 1 + nd_nx;
	       int n4 = n1 + nd_nx;

	       int n5 = n1 + nd_nxny;
	       int n6 = n1 + nd_nxny + 1;
	       int n7 = n1 + nd_nxny + 1 + nd_nx;
	       int n8 = n1 + nd_nxny + nd_nx;

               real8 vol = UpwindVolume( x[n1], x[n2], x[n3], x[n4],
                                         x[n5], x[n6], x[n7], x[n8],

                                         y[n1], y[n2], y[n3], y[n4],
                                         y[n5], y[n6], y[n7], y[n8],

                                         z[n1], z[n2], z[n3], z[n4],
                                         z[n5], z[n6], z[n7], z[n8] );

               cvol[cind] = vol;

	       real8 xmid = 0.125*( x[n1] + x[n2] + x[n3] + x[n4] +
				    x[n5] + x[n6] + x[n7] + x[n8] );

	       real8 ymid = 0.125*( y[n1] + y[n2] + y[n3] + y[n4] +
				    y[n5] + y[n6] + y[n7] + y[n8] );

	       real8 zmid = 0.125*( z[n1] + z[n2] + z[n3] + z[n4] +
				    z[n5] + z[n6] + z[n7] + z[n8] );


               real8 xc = xmid - d_center[0];
               real8 yc = ymid - d_center[0];
               real8 zc = zmid - d_center[0];

               real8 radsq = xc*xc + yc*yc + zc*zc;

               bool inside = (d_radius*d_radius) > radsq;

               if (inside) {
                  psi[cind] = d_uval_inside;
               } else {
                  psi[cind] = d_uval_outside;
               }

	    }
	 }
      }

   }

   if (d_use_nonuniform_workload) {
      if (!patch.checkAllocated(d_workload_data_id)) {
         patch.allocatePatchData(d_workload_data_id);
      }
      tbox::Pointer< pdat::CellData<NDIM,double> > workload_data =
         patch.getPatchData(d_workload_data_id);
      workload_data->fillAll(1.0);
   }

}

//
// ==================================================================
//    extra code
// ==================================================================
//




/*
*************************************************************************
*                                                                       *
* Compute stable time increment for patch.  Return this value.          *
*                                                                       *
*************************************************************************
*/

double MblkLinAdv::computeStableDtOnPatch(
   hier::Patch<NDIM>& patch,
   const bool initial_time,
   const double dt_time)
{
   (void) initial_time;
   (void) dt_time;


   /*
    * Build the mapped grid on the patch.
    */
   int block_number = getBlockNumber();
   int level_number =  patch.getPatchLevelNumber();
   setMappedGridOnPatch(patch, level_number, block_number);

   const hier::Index<NDIM> ifirst=patch.getBox().lower();
   const hier::Index<NDIM> ilast =patch.getBox().upper();

   tbox::Pointer< pdat::CellData<NDIM,double> > uval =
      patch.getPatchData(d_uval, getDataContext());
   tbox::Pointer< pdat::CellData<NDIM,double> > vol =
      patch.getPatchData(d_vol, getDataContext());
   tbox::Pointer< pdat::NodeData<NDIM,double> > xyz =
      patch.getPatchData(d_xyz, getDataContext());


#ifdef DEBUG_CHECK_ASSERTIONS
      TBOX_ASSERT(!uval.isNull());
      TBOX_ASSERT(!vol.isNull());
      TBOX_ASSERT(!xyz.isNull());
      TBOX_ASSERT(uval->getGhostCellWidth() == vol->getGhostCellWidth());
#endif

   hier::IntVector<NDIM> uval_ghosts = uval->getGhostCellWidth();
   hier::IntVector<NDIM> xyz_ghosts = xyz->getGhostCellWidth();

   /*
    * Adjust advection velocity based on rotation of block
    */
   //int block_number = getBlockNumber();

   int imin = ifirst(0) - uval_ghosts(0);
   int imax = ilast(0)  + uval_ghosts(0);
   int jmin = ifirst(1) - uval_ghosts(1);
   int jmax = ilast(1)  + uval_ghosts(1);
   int kmin = ifirst(2) - uval_ghosts(2);
   //int kmax = ilast(2)  + uval_ghosts(2);
   int nx   = imax - imin + 1;
   int ny   = jmax - jmin + 1;
   // int nz   = kmax - kmin + 1;
   int nxny = nx*ny;
   // int nel  = nx*ny*nz;

   int nd_imin = ifirst(0)     - xyz_ghosts(0);
   int nd_imax = ilast(0)  + 1 + xyz_ghosts(0);
   int nd_jmin = ifirst(1)     - xyz_ghosts(1);
   int nd_jmax = ilast(1)  + 1 + xyz_ghosts(1);
   int nd_kmin = ifirst(2)     - xyz_ghosts(2);
   //int nd_kmax = ilast(2)  + 1 + xyz_ghosts(2);
   int nd_nx   = nd_imax - nd_imin + 1;
   int nd_ny   = nd_jmax - nd_jmin + 1;
   // int nd_nz   = nd_kmax - nd_kmin + 1;
   int nd_nxny = nd_nx*nd_ny;
   // int nd_nel  = nd_nx*nd_ny*nd_nz;

   //int rot = d_mblk_geometry->getBlockRotation(block_number);

   real8 u = d_advection_velocity[0];
   real8 v = d_advection_velocity[1];
   real8 w = d_advection_velocity[2];

   //
   // compute the source due to the upwind method
   //
   double stabdt = 1.e20;;

   double *cvol = vol->getPointer();

   double *x = xyz->getPointer(0);
   double *y = xyz->getPointer(1);
   double *z = xyz->getPointer(2);

   for ( int k = ifirst(2); k <= ilast(2); k++ ) {
      for ( int j = ifirst(1); j <= ilast(1); j++ ) {
         for ( int i = ifirst(0); i <= ilast(0); i++ ) {

            int cind  = POLY3(i,j,k,imin,jmin,kmin,nx,nxny);

	    //   have set up the normal so that it always points in the positive, i, j, and k directions

            int n1 = POLY3(i,j,k,nd_imin,nd_jmin,nd_kmin,nd_nx,nd_nxny);
            int n2 = n1 + 1;
            int n3 = n1 + 1 + nd_nx;
            int n4 = n1 + nd_nx;

            int n5 = n1 + nd_nxny;
            int n6 = n1 + nd_nxny + 1;
            int n7 = n1 + nd_nxny + 1 + nd_nx;
            int n8 = n1 + nd_nxny + nd_nx;

            real8 vol = UpwindVolume( x[n1], x[n2], x[n3], x[n4],
				      x[n5], x[n6], x[n7], x[n8],

				      y[n1], y[n2], y[n3], y[n4],
				      y[n5], y[n6], y[n7], y[n8],

				      z[n1], z[n2], z[n3], z[n4],
				      z[n5], z[n6], z[n7], z[n8] );

            cvol[cind] = vol;

            if (vol < 0.) {
               TBOX_ERROR("Error:  negative volume computed in UpwindVolume");
            }

	    real8 xx[8];
	    real8 yy[8];
	    real8 zz[8];

	    xx[0] = x[n1];
	    xx[1] = x[n2];
	    xx[2] = x[n3];
	    xx[3] = x[n4];
	    xx[4] = x[n5];
	    xx[5] = x[n6];
	    xx[6] = x[n7];
	    xx[7] = x[n8];

	    yy[0] = y[n1];
	    yy[1] = y[n2];
	    yy[2] = y[n3];
	    yy[3] = y[n4];
	    yy[4] = y[n5];
	    yy[5] = y[n6];
	    yy[6] = y[n7];
	    yy[7] = y[n8];

	    zz[0] = z[n1];
	    zz[1] = z[n2];
	    zz[2] = z[n3];
	    zz[3] = z[n4];
	    zz[4] = z[n5];
	    zz[5] = z[n6];
	    zz[6] = z[n7];
	    zz[7] = z[n8];

	    real8 arealg = UpwindCharacteristicLength( xx, yy, zz, vol );

	    real8 uu = MAX( u, MAX( v, w ) );

	    stabdt = MIN( stabdt, arealg/uu );

	 }
      }
   }
   return stabdt;
}

/*
*************************************************************************
*                                                                       *
* Compute time integral of numerical fluxes for finite difference       *
* at each cell face on patch.  When NDIM == 3, there are two options    *
* for the transverse flux correction.  Otherwise, there is only one.    *
*                                                                       *
*************************************************************************
*/

void MblkLinAdv::computeFluxesOnPatch(hier::Patch<NDIM>& patch,
                                  const double time,
                                  const double dt)
{
   (void) time;

   //return;

   int block_number = getBlockNumber();
   int level_number =  patch.getPatchLevelNumber();
   setMappedGridOnPatch(patch, level_number, block_number);

   hier::Box<NDIM> pbox = patch.getBox();
   const hier::Index<NDIM> ifirst = patch.getBox().lower();
   const hier::Index<NDIM> ilast  = patch.getBox().upper();

   tbox::Pointer< pdat::CellData<NDIM,double> > uval =
      patch.getPatchData(d_uval, getDataContext());

   tbox::Pointer< pdat::CellData<NDIM,double> > vol =
      patch.getPatchData(d_vol, getDataContext());

   tbox::Pointer< pdat::SideData<NDIM,double> > flux =
      patch.getPatchData(d_flux, getDataContext());

   tbox::Pointer< pdat::NodeData<NDIM,double> > xyz =
      patch.getPatchData(d_xyz, getDataContext());

   /*
    * Verify that the integrator providing the context correctly
    * created it, and that the ghost cell width associated with the
    * context matches the ghosts defined in this class...
    */
#ifdef DEBUG_CHECK_ASSERTIONS
   TBOX_ASSERT(!uval.isNull());
   TBOX_ASSERT(!flux.isNull());
   TBOX_ASSERT(!vol.isNull());
   TBOX_ASSERT(uval->getGhostCellWidth() == d_nghosts);
   TBOX_ASSERT(vol->getGhostCellWidth() == d_nghosts);
   TBOX_ASSERT(flux->getGhostCellWidth() == d_fluxghosts);
   TBOX_ASSERT(xyz->getGhostCellWidth() == d_nodeghosts);
#endif

   //
   // ------------------------------- spliced in code ----------------------------
   //

   int fx_imn  = ifirst(0)     - d_fluxghosts(0);
   int fx_imx  = ilast(0)  + 1 + d_fluxghosts(0);
   int fx_jmn  = ifirst(1)     - d_fluxghosts(1);
   int fx_jmx  = ilast(1)      + d_fluxghosts(1);
   int fx_kmn  = ifirst(2)     - d_fluxghosts(2);
   //int fx_kmx  = ilast(2)      + d_fluxghosts(2);
   int fx_nx   = fx_imx - fx_imn + 1;
   int fx_ny   = fx_jmx - fx_jmn + 1;
   //int fx_nz   = fx_kmx - fx_kmn + 1;
   int fx_nxny = fx_nx*fx_ny;
   //int fx_nel  = fx_nx*fx_ny*fx_nz;

   int fy_imn  = ifirst(0)     - d_fluxghosts(0);
   int fy_imx  = ilast(0)      + d_fluxghosts(0);
   int fy_jmn  = ifirst(1)     - d_fluxghosts(1);
   int fy_jmx  = ilast(1)  + 1 + d_fluxghosts(1);
   int fy_kmn  = ifirst(2)     - d_fluxghosts(2);
   //int fy_kmx  = ilast(2)      + d_fluxghosts(2);
   int fy_nx   = fy_imx - fy_imn + 1;
   int fy_ny   = fy_jmx - fy_jmn + 1;
   //int fy_nz   = fy_kmx - fy_kmn + 1;
   int fy_nxny = fy_nx*fy_ny;
   //int fy_nel  = fy_nx*fy_ny*fy_nz;

   int fz_imn  = ifirst(0)      - d_fluxghosts(0);
   int fz_imx  = ilast(0)       + d_fluxghosts(0);
   int fz_jmn  = ifirst(1)      - d_fluxghosts(1);
   int fz_jmx  = ilast(1)       + d_fluxghosts(1);
   int fz_kmn  = ifirst(2)      - d_fluxghosts(2);
   //int fz_kmx  = ilast(2)   + 1 + d_fluxghosts(2);
   int fz_nx   = fz_imx - fz_imn + 1;
   int fz_ny   = fz_jmx - fz_jmn + 1;
   //int fz_nz   = fz_kmx - fz_kmn + 1;
   int fz_nxny = fz_nx*fz_ny;
   //int fz_nel  = fz_nx*fz_ny*fz_nz;

   int imin = ifirst(0) - d_nghosts(0);
   int imax = ilast(0)  + d_nghosts(0);
   int jmin = ifirst(1) - d_nghosts(1);
   int jmax = ilast(1)  + d_nghosts(1);
   int kmin = ifirst(2) - d_nghosts(2);
   //int kmax = ilast(2)  + d_nghosts(2);
   int nx   = imax - imin + 1;
   int ny   = jmax - jmin + 1;
   // int nz   = kmax - kmin + 1;
   int nxny = nx*ny;
   // int nel  = nx*ny*nz;

   int nd_imin = ifirst(0)     - d_nodeghosts(0);
   int nd_imax = ilast(0)  + 1 + d_nodeghosts(0);
   int nd_jmin = ifirst(1)     - d_nodeghosts(1);
   int nd_jmax = ilast(1)  + 1 + d_nodeghosts(1);
   int nd_kmin = ifirst(2)     - d_nodeghosts(2);
   //int nd_kmax = ilast(2)  + 1 + d_nodeghosts(2);
   int nd_nx   = nd_imax - nd_imin + 1;
   int nd_ny   = nd_jmax - nd_jmin + 1;
   // int nd_nz   = nd_kmax - nd_kmin + 1;
   int nd_nxny = nd_nx*nd_ny;
   // int nd_nel  = nd_nx*nd_ny*nd_nz;

   //
   // get the pointers
   //
   double *psi = uval->getPointer();

   double *cvol = vol->getPointer();

   double *fx = flux->getPointer(0);
   double *fy = flux->getPointer(1);
   double *fz = flux->getPointer(2);

   double *x = xyz->getPointer(0);
   double *y = xyz->getPointer(1);
   double *z = xyz->getPointer(2);

   real8 u = -d_advection_velocity[0];
   real8 v = -d_advection_velocity[1];
   real8 w = -d_advection_velocity[2];

   //
   // compute the source due to the upwind method
   //
   for ( int k = ifirst(2); k <= ilast(2); k++ ) {
      for ( int j = ifirst(1); j <= ilast(1); j++ ) {
         for ( int i = ifirst(0); i <= ilast(0)+1; i++ ) {

	    int ifx   = POLY3(i,j,k,fx_imn,fx_jmn,fx_kmn,fx_nx,fx_nxny);

	    // --------- get the neighbors
            int ind  = POLY3(i,j,k,imin,jmin,kmin,nx,nxny);
            int ib   = ind-1;

	    // ---------- use a righthand rule in evaluating face data, n1 -- n4 go around the bottom plane
	    //   of the element, n5-n8 go around the top plane of the element.  note, I
	    //   have set up the normal so that it always points in the positive, i, j, and k directions
            int n1 = POLY3(i,j,k,nd_imin,nd_jmin,nd_kmin,nd_nx,nd_nxny);
            int n4 = n1 + nd_nx;
            int n5 = n1 + nd_nxny;
            int n8 = n1 + nd_nxny + nd_nx;

	    fx[ifx] = UpwindFlux( x[n1], x[n4], x[n8], x[n5], // 1 - 4 - 8 - 5
				  y[n1], y[n4], y[n8], y[n5],
				  z[n1], z[n4], z[n8], z[n5],
				  u, v, w,
				  psi[ib], psi[ind] );
	 }
      }
   }


   for ( int k = ifirst(2); k <= ilast(2); k++ ) {
      for ( int j = ifirst(1); j <= ilast(1)+1; j++ ) {
         for ( int i = ifirst(0); i <= ilast(0); i++ ) {

	    int ify   = POLY3(i,j,k,fy_imn,fy_jmn,fy_kmn,fy_nx,fy_nxny);

	    // --------- get the neighbors
            int ind  = POLY3(i,j,k,imin,jmin,kmin,nx,nxny);
            int jb   = ind-nx;

	    // ---------- use a righthand rule in evaluating face data, n1 -- n4 go around the bottom plane
	    //   of the element, n5-n8 go around the top plane of the element.  note, I
	    //   have set up the normal so that it always points in the positive, i, j, and k directions
            int n1 = POLY3(i,j,k,nd_imin,nd_jmin,nd_kmin,nd_nx,nd_nxny);
            int n2 = n1 + 1;
            int n5 = n1 + nd_nxny;
            int n6 = n1 + nd_nxny + 1;

	    fy[ify] = UpwindFlux( x[n1], x[n5], x[n6], x[n2], // 1 - 5 - 6 - 2
				  y[n1], y[n5], y[n6], y[n2],
				  z[n1], z[n5], z[n6], z[n2],
				  u, v, w,
				  psi[jb], psi[ind] );
	 }
      }
   }

   for ( int k = ifirst(2); k <= ilast(2)+1; k++ ) {
      for ( int j = ifirst(1); j <= ilast(1); j++ ) {
         for ( int i = ifirst(0); i <= ilast(0); i++ ) {

	    int ifz   = POLY3(i,j,k,fz_imn,fz_jmn,fz_kmn,fz_nx,fz_nxny);

	    // --------- get the neighbors
            int ind  = POLY3(i,j,k,imin,jmin,kmin,nx,nxny);
            int kb   = ind-nxny;

	    // ---------- use a righthand rule in evaluating face data, n1 -- n4 go around the bottom plane
	    //   of the element, n5-n8 go around the top plane of the element.  note, I
	    //   have set up the normal so that it always points in the positive, i, j, and k directions
            int n1 = POLY3(i,j,k,nd_imin,nd_jmin,nd_kmin,nd_nx,nd_nxny);
            int n2 = n1 + 1;
            int n3 = n1 + 1 + nd_nx;
            int n4 = n1 + nd_nx;

	    fz[ifz] = UpwindFlux( x[n1], x[n2], x[n3], x[n4], // 1 - 2 - 3 - 4
				  y[n1], y[n2], y[n3], y[n4],
				  z[n1], z[n2], z[n3], z[n4],
				  u, v, w,
				  psi[kb], psi[ind] );
	 }
      }
   }



   //
   // compute the source due to the upwind method
   //
   for ( int k = ifirst(2); k <= ilast(2); k++ ) {
      for ( int j = ifirst(1); j <= ilast(1); j++ ) {
         for ( int i = ifirst(0); i <= ilast(0); i++ ) {

            int ind  = POLY3(i,j,k,imin,jmin,kmin,nx,nxny);

	    int ib   = POLY3(i,  j,k,fx_imn,fx_jmn,fx_kmn,fx_nx,fx_nxny);
	    int ie   = POLY3(i+1,j,k,fx_imn,fx_jmn,fx_kmn,fx_nx,fx_nxny);

	    int jb   = POLY3(i,j,  k,fy_imn,fy_jmn,fy_kmn,fy_nx,fy_nxny);
	    int je   = POLY3(i,j+1,k,fy_imn,fy_jmn,fy_kmn,fy_nx,fy_nxny);

	    int kb   = POLY3(i,j,k,  fz_imn,fz_jmn,fz_kmn,fz_nx,fz_nxny);
	    int ke   = POLY3(i,j,k+1,fz_imn,fz_jmn,fz_kmn,fz_nx,fz_nxny);


	    //   have set up the normal so that it always points in the positive, i, j, and k directions
            int n1 = POLY3(i,j,k,nd_imin,nd_jmin,nd_kmin,nd_nx,nd_nxny);
            int n2 = n1 + 1;
            int n3 = n1 + 1 + nd_nx;
            int n4 = n1 + nd_nx;

            int n5 = n1 + nd_nxny;
            int n6 = n1 + nd_nxny + 1;
            int n7 = n1 + nd_nxny + 1 + nd_nx;
            int n8 = n1 + nd_nxny + nd_nx;

            real8 vol = UpwindVolume( x[n1], x[n2], x[n3], x[n4],
				      x[n5], x[n6], x[n7], x[n8],

				      y[n1], y[n2], y[n3], y[n4],
				      y[n5], y[n6], y[n7], y[n8],

				      z[n1], z[n2], z[n3], z[n4],
				      z[n5], z[n6], z[n7], z[n8] );

            cvol[ind] = vol;


	    psi[ind] -= dt*( ( fx[ie] - fx[ib] ) +
            		     ( fy[je] - fy[jb] ) +
            		     ( fz[ke] - fz[kb] ) )/vol;


	 }
      }
   }



}


/*
*************************************************************************
*                                                                       *
* Update solution variables by performing a conservative                *
* difference with the fluxes calculated in computeFluxesOnPatch().      *
*                                                                       *
*************************************************************************
*/

void MblkLinAdv::conservativeDifferenceOnPatch(hier::Patch<NDIM>& patch,
                                           const double time,
                                           const double dt,
                                           bool at_syncronization)
{
   (void) time;
   (void) dt;
   (void) at_syncronization;

   return;
}

/*
*************************************************************************
*                                                                       *
* Set the data in ghost cells corresponding to physical boundary        *
* conditions.  Note that boundary geometry configuration information    *
* (i.e., faces, edges, and nodes) is obtained from the patch geometry   *
* object owned by the patch.                                            *
*                                                                       *
*************************************************************************
*/

void MblkLinAdv::setPhysicalBoundaryConditions(
   hier::Patch<NDIM>& patch,
   const double fill_time,
   const hier::IntVector<NDIM>& ghost_width_to_fill)
{
   (void) fill_time;

   tbox::Pointer< pdat::CellData<NDIM,double> > uval =
      patch.getPatchData(d_uval, getDataContext());

#ifdef DEBUG_CHECK_ASSERTIONS
   TBOX_ASSERT(!uval.isNull());
#endif
   hier::IntVector<NDIM> uval_ghosts = uval->getGhostCellWidth();
#ifdef DEBUG_CHECK_ASSERTIONS
   TBOX_ASSERT(uval->getGhostCellWidth() == d_nghosts);
#endif

#if (NDIM == 2)

   /*
    * Set boundary conditions for cells corresponding to patch edges.
    */
   SkeletonBoundaryUtilities2::
      fillEdgeBoundaryData("uval", uval,
                           patch,
                           ghost_width_to_fill,
                           d_scalar_bdry_edge_conds,
                           d_bdry_edge_uval);


   /*
    *  Set boundary conditions for cells corresponding to patch nodes.
    */

   SkeletonBoundaryUtilities2::
      fillNodeBoundaryData("uval", uval,
                           patch,
                           ghost_width_to_fill,
                           d_scalar_bdry_node_conds,
                           d_bdry_edge_uval);

#endif // NDIM == 2

#if (NDIM == 3)

   /*
    *  Set boundary conditions for cells corresponding to patch faces.
    */

   SkeletonBoundaryUtilities3::
      fillFaceBoundaryData("uval", uval,
                           patch,
                           ghost_width_to_fill,
                           d_scalar_bdry_face_conds,
                           d_bdry_face_uval);

   /*
    *  Set boundary conditions for cells corresponding to patch edges.
    */

   SkeletonBoundaryUtilities3::
      fillEdgeBoundaryData("uval", uval,
                           patch,
                           ghost_width_to_fill,
                           d_scalar_bdry_edge_conds,
                           d_bdry_face_uval);

   /*
    *  Set boundary conditions for cells corresponding to patch nodes.
    */

   SkeletonBoundaryUtilities3::
      fillNodeBoundaryData("uval", uval,
                           patch,
                           ghost_width_to_fill,
                           d_scalar_bdry_node_conds,
                           d_bdry_face_uval);
#endif // NDIM == 3

}


/*
*************************************************************************
*                                                                       *
* Refine operations
*                                                                       *
*************************************************************************
*/

void MblkLinAdv::preprocessRefine(
   hier::Patch<NDIM>& fine,
   const hier::Patch<NDIM>& coarse,
   const hier::Box<NDIM>& fine_box,
   const hier::IntVector<NDIM>& ratio )
{

   int xyz_id =  hier::VariableDatabase<NDIM>::getDatabase()->
      mapVariableAndContextToIndex(d_xyz, getDataContext());

   int fln = fine.getPatchLevelNumber();
   int cln = coarse.getPatchLevelNumber();
   if (fln < 0) {
      fln = cln + 1;
      if (!fine.checkAllocated(xyz_id)) {
         TBOX_ERROR(d_object_name << ":preprocessRefine()"
                    << "\nfine xyz data not allocated" << endl);
      }
   }
   if (cln < 0) {
      cln = fln - 1;
      if (!coarse.checkAllocated(xyz_id)) {
         TBOX_ERROR(d_object_name << ":preprocessRefine()"
                    << "\ncoarse xyz data not allocated" << endl);
      }
   }
   int block_number = getBlockNumber();
   setMappedGridOnPatch(coarse, cln, block_number);
   setMappedGridOnPatch(fine, fln, block_number);


}


void MblkLinAdv::postprocessRefine(
   hier::Patch<NDIM>& fine,
   const hier::Patch<NDIM>& coarse,
   const hier::Box<NDIM>& fine_box,
   const hier::IntVector<NDIM>& ratio )
{

   tbox::Pointer< pdat::CellData<NDIM,double> > cuval     =
      coarse.getPatchData(d_uval, getDataContext());
   tbox::Pointer< pdat::CellData<NDIM,double> > cvol     =
      coarse.getPatchData(d_vol, getDataContext());

   tbox::Pointer< pdat::CellData<NDIM,double> > fuval     =
      fine.getPatchData(d_uval, getDataContext());
   tbox::Pointer< pdat::CellData<NDIM,double> > fvol     =
      fine.getPatchData(d_vol, getDataContext());

#ifdef DEBUG_CHECK_ASSERTIONS
   TBOX_ASSERT(!cuval.isNull());
   TBOX_ASSERT(!fuval.isNull());
   TBOX_ASSERT(!cvol.isNull());
   TBOX_ASSERT(!fvol.isNull());
   TBOX_ASSERT(cuval->getDepth() == fuval->getDepth());
#endif

   //
   // get the needed geometry and box information
   //
   const hier::Box<NDIM> cgbox(cuval->getGhostBox() );

   const hier::Index<NDIM> cilo = cgbox.lower();
   const hier::Index<NDIM> cihi = cgbox.upper();
   const hier::Index<NDIM> filo = fuval->getGhostBox().lower();
   const hier::Index<NDIM> fihi = fuval->getGhostBox().upper();

   const hier::Box<NDIM> coarse_box = hier::Box<NDIM>::coarsen(fine_box, ratio );
   const hier::Index<NDIM> ifirstc = coarse_box.lower();
   const hier::Index<NDIM> ilastc  = coarse_box.upper();
   const hier::Index<NDIM> ifirstf = fine_box.lower();
   const hier::Index<NDIM> ilastf  = fine_box.upper();

   int flev = fine.getPatchLevelNumber();
   int clev = coarse.getPatchLevelNumber();

   tbox::plog << "--------------------- start postprocessRefineData (" << flev << "," << clev << ")" << endl;
   tbox::plog << "flevel     = " << flev << endl;
   tbox::plog << "clevel     = " << clev << endl;
   tbox::plog << "fine       = " << fine.getBox()   << endl;
   tbox::plog << "coarse     = " << coarse.getBox() << endl;
   tbox::plog << "fine_box   = " << fine_box        << endl;
   tbox::plog << "coarse_box = " << coarse_box      << endl;

   tbox::plog << "filo = " << filo << ", fihi = " << fihi << endl;
   tbox::plog << "cilo = " << cilo << ", cihi = " << cihi << endl;

   //
   // setup work variables
   //
   const hier::IntVector<NDIM> tmp_ghosts(0 );
   const int numInterp = 1;
   pdat::CellData<NDIM,double> val_vals(    coarse_box, numInterp, tmp_ghosts );
   pdat::CellData<NDIM,double> slope0_vals( coarse_box, numInterp, tmp_ghosts );
   pdat::CellData<NDIM,double> slope1_vals( coarse_box, numInterp, tmp_ghosts );
   pdat::CellData<NDIM,double> slope2_vals( coarse_box, numInterp, tmp_ghosts );
   double *val      	  = val_vals.getPointer();
   double *slope0 	  = slope0_vals.getPointer();
   double *slope1 	  = slope1_vals.getPointer();
   double *slope2 	  = slope2_vals.getPointer();


   int imin = ifirstc(0);  // the box of coarse elements being refined
   int imax = ilastc(0);   // the work data is sized to this box
   int jmin = ifirstc(1);
   int jmax = ilastc(1);
   int kmin = ifirstc(2);
   int kmax = ilastc(2);
   int nx   = imax - imin + 1;
   int ny   = jmax - jmin + 1;
   int nz   = kmax - kmin + 1;
   int nxny = nx*ny;
   int nel  = nx*ny*nz;

   int cimin = cilo(0);  // the coarse data bounds
   int cimax = cihi(0);
   int cjmin = cilo(1);
   int cjmax = cihi(1);
   int ckmin = cilo(2);
   int ckmax = cihi(2);
   int cnx   = cimax - cimin + 1;
   int cny   = cjmax - cjmin + 1;
   int cnz   = ckmax - ckmin + 1;
   int cnxny = cnx*cny;
   int cnel  = cnx*cny*cnz;

   int fimin = filo(0);  // the fine data bounds
   int fimax = fihi(0);
   int fjmin = filo(1);
   int fjmax = fihi(1);
   int fkmin = filo(2);
   int fkmax = fihi(2);
   int fnx   = fimax - fimin + 1;
   int fny   = fjmax - fjmin + 1;
   int fnz   = fkmax - fkmin + 1;
   int fnxny = fnx*fny;
   int fnel  = fnx*fny*fnz;

   double rat0 = ratio[0];
   double rat1 = ratio[1];
   double rat2 = ratio[2];
   double fact = 2.0;  // xi varies from -1 to 1

   double *cvolume        = cvol->getPointer();
   double *cdata          = cuval->getPointer();
   double *fdata          = fuval->getPointer();

   //
   // ================================= history variable refinement ====================
   //
   for ( int n = 0 ; n < DEPTH; n++ ) {

      for ( int l = 0; l < nel ; l++ ) {    // default slopes are zero
	 slope0[l] = 0.0;  // this yields piecewise constant interpolation
	 slope1[l] = 0.0;  // and makes a handy initializer
	 slope2[l] = 0.0;
      }

      for ( int k = ifirstc(2); k <= ilastc(2); k++ ) {
	 for ( int j = ifirstc(1); j <= ilastc(1); j++ ) {
	    for ( int i = ifirstc(0); i <= ilastc(0); i++ ) {

	       int ind   = POLY3(i,j,k,imin,jmin,kmin,nx,nxny);
	       int chind = POLY3(i,j,k,cimin,cjmin,ckmin,cnx,cnxny) + n*cnel;

	       int wind = POLY3(i,j,k,cimin,cjmin,ckmin,cnx,cnxny);
               double w_i  = cvolume[ wind       ];
               double w_im = cvolume[ wind-1     ];
               double w_ip = cvolume[ wind+1     ];
               double w_jm = cvolume[ wind-cnx   ];
               double w_jp = cvolume[ wind+cnx   ];
               double w_km = cvolume[ wind-cnxny ];
               double w_kp = cvolume[ wind+cnxny ];

	       int im1 = chind-1;
	       int ip1 = chind+1;
	       int jm1 = chind-cnx;
	       int jp1 = chind+cnx;
	       int km1 = chind-cnxny;
	       int kp1 = chind+cnxny;

	       double aii = cdata[ chind ];

	       double aip = cdata[ ip1 ];
	       double aim = cdata[ im1 ];
	       double ajp = cdata[ jp1 ];
	       double ajm = cdata[ jm1 ];
	       double akp = cdata[ kp1 ];
	       double akm = cdata[ km1 ];

	       val[ind] = aii;

#ifdef DEBUG_CHECK_ASSERTIONS
	       TBOX_ASSERT( ind >= 0 ); // debug assertions
	       TBOX_ASSERT( ind < nel );
#endif
	       my_slopes( aii, aip, aim, ajp, ajm, akp, akm,
                          w_i, w_ip, w_im, w_jp, w_jm, w_kp, w_km,
                          slope0[ind],
                          slope1[ind],
                          slope2[ind] );

	    }
	 }
      }

      //
      // compute the interpolated data from the cached slopes, looping
      // over the fine zones
      //
      for ( int k = ifirstf(2); k <= ilastf(2); k++ ) {
	 for ( int j = ifirstf(1); j <= ilastf(1); j++ ) {
	    for ( int i = ifirstf(0); i <= ilastf(0); i++ ) {

	       int fhind = POLY3(i,j,k,fimin,fjmin,fkmin,fnx,fnxny) + n*fnel;

	       double ric   = (double(i) + 0.5)/rat0 - 0.5;
	       double rjc   = (double(j) + 0.5)/rat1 - 0.5;
	       double rkc   = (double(k) + 0.5)/rat2 - 0.5;

	       int ic    = (int) (ric + ( ric >= 0 ? 0.5 : -0.5 ) );
	       int jc    = (int) (rjc + ( rjc >= 0 ? 0.5 : -0.5 ) );
	       int kc    = (int) (rkc + ( rkc >= 0 ? 0.5 : -0.5 ) );

	       double ldx   = fact*( ric - ic );
	       double ldy   = fact*( rjc - jc );
	       double ldz   = fact*( rkc - kc );

	       int ind  = POLY3(ic,jc,kc,imin,jmin,kmin,nx,nxny);  // work pos

#ifdef DEBUG_CHECK_ASSERTIONS
	       TBOX_ASSERT( 0   <= ind );
	       TBOX_ASSERT( ind <  nel );
#endif

	       fdata[fhind] =
		  cdata[ind] +
		  ldx*slope0[ind] +
		  ldy*slope1[ind] +
		  ldz*slope2[ind];

	    }
	 }
      } // end of i,j,k loops for finding the fine history variables

   } // end of history loop

   tbox::plog << "--------------------- end postprocessRefine" << endl;


}


/*
*************************************************************************
*                                                                       *
* Coarsen operations
*                                                                       *
*************************************************************************
*/
void MblkLinAdv::preprocessCoarsen(hier::Patch<NDIM>& coarse,
                                   const hier::Patch<NDIM>& fine,
                                   const hier::Box<NDIM>& coarse_box,
                                   const hier::IntVector<NDIM>& ratio)
{
   int xyz_id =  hier::VariableDatabase<NDIM>::getDatabase()->
      mapVariableAndContextToIndex(d_xyz, getDataContext());

   int fln = fine.getPatchLevelNumber();
   int cln = coarse.getPatchLevelNumber();
   if (fln < 0) {
      fln = cln + 1;
      if (!fine.checkAllocated(xyz_id)) {
         TBOX_ERROR(d_object_name << ":preprocessCoarsen()"
                    << "\nfine xyz data not allocated" << endl);
      }
   }
   if (cln < 0) {
      cln = fln - 1;
      if (!coarse.checkAllocated(xyz_id)) {
         TBOX_ERROR(d_object_name << ":preprocessCoarsen()"
                    << "\ncoarse xyz data not allocated" << endl);
      }
   }
   int block_number = getCoarsenBlockNumber();
   setMappedGridOnPatch(coarse, cln, block_number);
   setMappedGridOnPatch(fine, fln, block_number);

}

//
// the coarsening function
//
void MblkLinAdv::postprocessCoarsen( hier::Patch<NDIM>& coarse,
                                     const hier::Patch<NDIM>& fine,
                                     const hier::Box<NDIM>& coarse_box,
                                     const hier::IntVector<NDIM>& ratio )
{


   tbox::Pointer< pdat::CellData<NDIM,double> > cuval     =
      coarse.getPatchData(d_uval, getDataContext());
   tbox::Pointer< pdat::CellData<NDIM,double> > cvol      =
      coarse.getPatchData(d_vol, getDataContext());

   tbox::Pointer< pdat::CellData<NDIM,double> > fuval     =
      fine.getPatchData(d_uval, getDataContext());
   tbox::Pointer< pdat::CellData<NDIM,double> > fvol      =
      fine.getPatchData(d_vol, getDataContext());

#ifdef DEBUG_CHECK_ASSERTIONS
   TBOX_ASSERT(!cuval.isNull());
   TBOX_ASSERT(!cvol.isNull());
   TBOX_ASSERT(!fuval.isNull());
   TBOX_ASSERT(!fvol.isNull());
   TBOX_ASSERT(cuval->getDepth() == fuval->getDepth());
#endif

   //
   // box and geometry information
   //
   const hier::Index<NDIM> filo = fuval->getGhostBox().lower();
   const hier::Index<NDIM> fihi = fuval->getGhostBox().upper();
   const hier::Index<NDIM> cilo = cuval->getGhostBox().lower();
   const hier::Index<NDIM> cihi = cuval->getGhostBox().upper();

   const hier::Box<NDIM> fine_box = hier::Box<NDIM>::refine(coarse_box, ratio );
   const hier::Index<NDIM> ifirstc = coarse_box.lower();  // coarse basis
   const hier::Index<NDIM> ilastc  = coarse_box.upper();
   const hier::Index<NDIM> ifirstf = fine_box.lower();    // fine data
   const hier::Index<NDIM> ilastf  = fine_box.upper();

   int flev = fine.getPatchLevelNumber();
   int clev = coarse.getPatchLevelNumber();

   tbox::plog << "--------------------- start postprocessCoarsenData (" << flev << "," << clev << ")" << endl;
   tbox::plog << "flevel     = " << flev << endl;
   tbox::plog << "clevel     = " << clev << endl;
   tbox::plog << "fine       = " << fine.getBox()   << endl;
   tbox::plog << "coarse     = " << coarse.getBox() << endl;
   tbox::plog << "fine_box   = " << fine_box        << endl;
   tbox::plog << "coarse_box = " << coarse_box      << endl;

   tbox::plog << "filo = " << filo << ", fihi = " << fihi << endl;
   tbox::plog << "cilo = " << cilo << ", cihi = " << cihi << endl;

   //
   // work variables
   //

   int cimin = cilo(0);
   int cimax = cihi(0);
   int cjmin = cilo(1);
   int cjmax = cihi(1);
   int ckmin = cilo(2);
   int ckmax = cihi(2);
   int cnx   = cimax - cimin + 1;
   int cny   = cjmax - cjmin + 1;
   int cnz   = ckmax - ckmin + 1;
   int cnxny = cnx*cny;
   int cnel  = cnx*cny*cnz;

   int fimin = filo(0);
   int fimax = fihi(0);
   int fjmin = filo(1);
   int fjmax = fihi(1);
   int fkmin = filo(2);
   int fkmax = fihi(2);
   int fnx   = fimax - fimin + 1;
   int fny   = fjmax - fjmin + 1;
   int fnz   = fkmax - fkmin + 1;
   int fnxny = fnx*fny;
   int fnel  = fnx*fny*fnz;

   double rat0 = (double)(ratio(0));
   double rat1 = (double)(ratio(1));
   double rat2 = (double)(ratio(2));

   double *cvolume        = cvol->getPointer();
   double *fvolume        = fvol->getPointer();
   double *cdata          = cuval->getPointer();
   double *fdata          = fuval->getPointer();

   //
   // average the data
   //
   for ( int n = 0 ; n < DEPTH; n++ ) {

      //
      // zero out the underlying coarse data to serve as a counter
      //
      for ( int k = ifirstc(2); k <= ilastc(2); k++ ) {  // loop over the coarse zones
	 for ( int j = ifirstc(1); j <= ilastc(1); j++ ) {
	    for ( int i = ifirstc(0); i <= ilastc(0); i++ ) {

	       int chind     = POLY3(i,j,k,cimin,cjmin,ckmin,cnx,cnxny) + n*cnel;

	       cdata[chind] = 0.0;
	    }
	 }
      }

      //
      // compute the interpolated data from the cached slopes
      //

      for ( int k = ifirstf(2); k <= ilastf(2); k++ ) {  // loop over the coarse zones
	 for ( int j = ifirstf(1); j <= ilastf(1); j++ ) {
	    for ( int i = ifirstf(0); i <= ilastf(0); i++ ) {

	       int vol_ind = POLY3(i,j,k,fimin,fjmin,fkmin,fnx,fnxny);
	       int fhind   = vol_ind + n*fnel;

	       real8 ric   = (double(i) + 0.5)/rat0 - 0.5;
	       real8 rjc   = (double(j) + 0.5)/rat1 - 0.5;
	       real8 rkc   = (double(k) + 0.5)/rat2 - 0.5;

	       int ic    = (int) (ric + ( ric >= 0 ? 0.5 : -0.5 ) ); // a round operation
	       int jc    = (int) (rjc + ( rjc >= 0 ? 0.5 : -0.5 ) ); // shift up and truncate if ic > 0
	       int kc    = (int) (rkc + ( rkc >= 0 ? 0.5 : -0.5 ) ); // shift down and truncate if ic < 0

	       int chind  = POLY3(ic,jc,kc,cimin,cjmin,ckmin,cnx,cnxny) + n*cnel; // post + history offset

#ifdef DEBUG_CHECK_ASSERTIONS
	       TBOX_ASSERT( cimin <= ic && ic <= cimax );
	       TBOX_ASSERT( cjmin <= jc && jc <= cjmax );
	       TBOX_ASSERT( ckmin <= kc && kc <= ckmax );
#endif

	       real8 fmass = fvolume[vol_ind];

	       cdata[chind] += fmass*fdata[fhind];  // sum of fine extensives now
	    }
	 }
      }

      //
      // normalize the completed sum by converting back from extensive to intensive
      //
      for ( int k = ifirstc(2); k <= ilastc(2); k++ ) {  // loop over the coarse zones
	 for ( int j = ifirstc(1); j <= ilastc(1); j++ ) {
	    for ( int i = ifirstc(0); i <= ilastc(0); i++ ) {

	       int vol_ind = POLY3(i,j,k,cimin,cjmin,ckmin,cnx,cnxny);
	       int chind   = vol_ind + n*cnel;

	       real8 cmass = cvolume[vol_ind];

	       cdata[chind] /= cmass + 1.e-80;
	    }
	 }
      }

   }
   tbox::plog << "--------------------- end postprocessCoarsen" << endl;
}



// -------------------------------------------------------------------


/*
*************************************************************************
*                                                                       *
* Tag cells for refinement using gradient detector.  Tagging criteria   *
* defined in input.                                                     *
*                                                                       *
*************************************************************************
*/

//
// tag cells for refinement
//
void MblkLinAdv::tagGradientDetectorCells( hier::Patch<NDIM>& patch,
					   const double regrid_time,
					   const bool initial_error,
					   const int tag_indx,
					   const bool uses_richardson_extrapolation_too)
{


   int block_number = getBlockNumber();
   int level_number =  patch.getPatchLevelNumber();
   setMappedGridOnPatch(patch, level_number, block_number);


   //
   // get geometry data
   //
   const int error_level_number = patch.getPatchLevelNumber();

   tbox::Pointer< pdat::NodeData<NDIM,double> > xyz =
      patch.getPatchData( d_xyz, getDataContext() );
   double *x = xyz->getPointer(0);
   double *y = xyz->getPointer(1);
   double *z = xyz->getPointer(2);

   hier::Box<NDIM> pbox           = patch.getBox();
   const hier::Index<NDIM> ifirst = patch.getBox().lower();
   const hier::Index<NDIM> ilast  = patch.getBox().upper();
   int level                = patch.getPatchLevelNumber();

   tbox::plog << "--------------------- start tagGradientCells (" << level << ")" << endl;
   tbox::plog << "level  = " << level << endl;
   tbox::plog << "box    = " << patch.getBox() << endl;

   tbox::Pointer< pdat::CellData<NDIM,int> > tags =
      patch.getPatchData(tag_indx );
   tbox::Pointer< pdat::CellData<NDIM, double > > var =
      patch.getPatchData( d_uval, getDataContext() );

   //
   // Create a set of temporary tags and set to untagged value.
   //
   tbox::Pointer< pdat::CellData<NDIM,int> > temp_tags  =
      new pdat::CellData<NDIM,int>(pbox, 1, d_nghosts );
   temp_tags->fillAll(FALSE );

   hier::IntVector<NDIM> tag_ghost = tags->getGhostCellWidth();

   hier::IntVector<NDIM> nghost_cells = xyz->getGhostCellWidth();
   int nd_imin = ifirst(0)     - nghost_cells(0);
   int nd_imax = ilast(0)  + 1 + nghost_cells(0);
   int nd_jmin = ifirst(1)     - nghost_cells(1);
   int nd_jmax = ilast(1)  + 1 + nghost_cells(1);
   int nd_kmin = ifirst(2)     - nghost_cells(2);
   //int nd_kmax = ilast(2)  + 1 + nghost_cells(2);
   int nd_nx   = nd_imax - nd_imin + 1;
   int nd_ny   = nd_jmax - nd_jmin + 1;
   //int nd_nz   = nd_kmax - nd_kmin + 1;
   int nd_nxny = nd_nx*nd_ny;
   //int nd_nel  = nd_nx*nd_ny*nd_nz;

   hier::IntVector<NDIM> v_ghost = var->getGhostCellWidth();     // has ghost zones
   int imin = ifirst(0) - v_ghost(0); // the polynomial for the field
   int imax = ilast(0)  + v_ghost(0);
   int jmin = ifirst(1) - v_ghost(1);
   int jmax = ilast(1)  + v_ghost(1);
   int kmin = ifirst(2) - v_ghost(2);
   //int kmax = ilast(2)  + v_ghost(2);
   int nx   = imax - imin + 1;
   int ny   = jmax - jmin + 1;
   // int nz   = kmax - kmin + 1;
   int nxny = nx*ny;
   //int nel  = nx*ny*nz;

   hier::IntVector<NDIM> temp_tag_ghost = temp_tags->getGhostCellWidth();
   int imn = ifirst(0) - temp_tag_ghost(0);  // the polynomial for temp_tags
   int imx = ilast(0)  + temp_tag_ghost(0);
   int jmn = ifirst(1) - temp_tag_ghost(1);
   int jmx = ilast(1)  + temp_tag_ghost(1);
   int kmn = ifirst(2) - temp_tag_ghost(2);
   //int kmx = ilast(2)  + temp_tag_ghost(2);
   int tnx   = imx - imn + 1;
   int tny   = jmx - jmn + 1;
   //int tnz   = kmx - kmn + 1;
   int tnxny = tnx*tny;
   //int tnel  = tnx*tny*tnz;

   double *lvar = var->getPointer();
   int *ltags   = temp_tags->getPointer();
   double  dv_x[3];
   double dv_xi[3];
   double xfact = 0.5;


   //
   // Possible tagging criteria includes
   //    DENSITY_DEVIATION, DENSITY_GRADIENT, DENSITY_SHOCK
   //    PRESSURE_DEVIATION, PRESSURE_GRADIENT, PRESSURE_SHOCK
   // The criteria is specified over a time interval.
   //
   // Loop over criteria provided and check to make sure we are in the
   // specified time interval.  If so, apply appropriate tagging for
   // the level.
   //
   for (int ncrit = 0; ncrit < d_refinement_criteria.getSize(); ncrit++) {

#ifdef DEBUG_CHECK_ASSERTIONS
      TBOX_ASSERT( !var.isNull() );
#endif
      string ref = d_refinement_criteria[ncrit];
      int size      = 0;
      double tol    = 0.;

      if (ref == "UVAL_GRADIENT" ) {
         size = d_grad_tol.getSize();  // max depth of gradient tolerance
         tol = ( ( error_level_number < size)  // find the tolerance
                 ? d_grad_tol[error_level_number]
                 : d_grad_tol[size-1]  );

         for ( int k = ifirst(2) ; k <= ilast(2) ; k++ ) {
            for ( int j = ifirst(1) ; j <= ilast(1) ; j++ ) {
               for ( int i = ifirst(0) ; i <= ilast(0) ; i++ ) {

                  int  ind = POLY3(i,j,k,imin,jmin,kmin,nx,nxny);
                  int tind = POLY3(i,j,k,imn,jmn,kmn,tnx,tnxny);

                  int ib = ind-1;
                  int ie = ind+1;
                  int jb = ind-nx;
                  int je = ind+nx;
                  int kb = ind-nxny;
                  int ke = ind+nxny;

                  //
                  // vector is now a max gradient in xi, eta, zeta
                  //
                  dv_xi[0] = xfact * MAX( fabs( lvar[ind]-lvar[ib] ), fabs( lvar[ind]-lvar[ie] ) );
                  dv_xi[1] = xfact * MAX( fabs( lvar[ind]-lvar[jb] ), fabs( lvar[ind]-lvar[je] ) );
                  dv_xi[2] = xfact * MAX( fabs( lvar[ind]-lvar[kb] ), fabs( lvar[ind]-lvar[ke] ) );

                  //
                  // convert vector to max gradine in x, y, z
                  //
                  int n1 = POLY3(i,j,k,nd_imin,nd_jmin,nd_kmin,nd_nx,nd_nxny);  // -1, -1, -1
                  int n2 = n1 + 1;                                              //  1, -1, -1
                  int n3 = n1 + 1 + nd_nx;                                      //  1,  1, -1
                  int n4 = n1 + nd_nx;                                          // -1,  1, -1

                  int n5 = n1 + nd_nxny;                                        // -1, -1,  1
                  int n6 = n1 + nd_nxny + 1;		   	             //  1, -1,  1
                  int n7 = n1 + nd_nxny + 1 + nd_nx;			     //  1,  1,  1
                  int n8 = n1 + nd_nxny + nd_nx;	         		     // -1,  1,  1

                  // ------------------------------------------------ x

                  real8 x1 = 0.25*( x[n1] + x[n4] + x[n5] + x[n8] );  // xi
                  real8 x2 = 0.25*( x[n2] + x[n3] + x[n6] + x[n7] );

                  real8 x3 = 0.25*( x[n1] + x[n2] + x[n5] + x[n6] );  // eta
                  real8 x4 = 0.25*( x[n3] + x[n4] + x[n7] + x[n8] );

                  real8 x5 = 0.25*( x[n1] + x[n2] + x[n3] + x[n4] );  // zeta
                  real8 x6 = 0.25*( x[n5] + x[n6] + x[n7] + x[n8] );

                  // ------------------------------------------------ y

                  real8 y1 = 0.25*( y[n1] + y[n4] + y[n5] + y[n8] );
                  real8 y2 = 0.25*( y[n2] + y[n3] + y[n6] + y[n7] );

                  real8 y3 = 0.25*( y[n1] + y[n2] + y[n5] + y[n6] );
                  real8 y4 = 0.25*( y[n3] + y[n4] + y[n7] + y[n8] );

                  real8 y5 = 0.25*( y[n1] + y[n2] + y[n3] + y[n4] );
                  real8 y6 = 0.25*( y[n5] + y[n6] + y[n7] + y[n8] );

                  // ------------------------------------------------ z

                  real8 z1 = 0.25*( z[n1] + z[n4] + z[n5] + z[n8] );
                  real8 z2 = 0.25*( z[n2] + z[n3] + z[n6] + z[n7] );

                  real8 z3 = 0.25*( z[n1] + z[n2] + z[n5] + z[n6] );
                  real8 z4 = 0.25*( z[n3] + z[n4] + z[n7] + z[n8] );

                  real8 z5 = 0.25*( z[n1] + z[n2] + z[n3] + z[n4] );
                  real8 z6 = 0.25*( z[n5] + z[n6] + z[n7] + z[n8] );

                  //
                  // the components of the matrices that we want to invert
                  //
                  real8 dx_xi   = 0.5*(x2 - x1);
                  real8 dy_xi   = 0.5*(y2 - y1);
                  real8 dz_xi   = 0.5*(z2 - z1);

                  real8 dx_eta  = 0.5*(x4 - x3);
                  real8 dy_eta  = 0.5*(y4 - y3);
                  real8 dz_eta  = 0.5*(z4 - z3);

                  real8 dx_zeta = 0.5*(x6 - x5);
                  real8 dy_zeta = 0.5*(y6 - y5);
                  real8 dz_zeta = 0.5*(z6 - z5);

                  //
                  // invert dx/dxi as in dx/dxi d/dx = d/dxi, note this
                  // is the transpose of the above matrix M, also via
                  // Kramer's rule
                  //
                  real8 detMt = ( dx_xi   * dy_eta  * dz_zeta +
                                  dx_eta  * dy_zeta * dz_xi   +
                                  dx_zeta * dy_xi   * dz_eta  -
                                  dx_zeta * dy_eta  * dz_xi   -
                                  dx_eta  * dy_xi   * dz_zeta -
                                  dx_xi   * dy_zeta * dz_eta  );

                  real8 detC11 = dy_eta  * dz_zeta - dz_eta  * dy_zeta;
                  real8 detC21 = dx_eta  * dz_zeta - dz_eta  * dx_zeta;
                  real8 detC31 = dx_eta  * dy_zeta - dy_eta  * dx_zeta;

                  real8 detC12 = dy_xi   * dz_zeta - dz_xi   * dy_zeta;
                  real8 detC22 = dx_xi   * dz_zeta - dz_xi   * dx_zeta;
                  real8 detC32 = dx_xi   * dy_zeta - dy_xi   * dx_zeta;

                  real8 detC13 = dy_xi   * dz_eta  - dz_xi   * dy_eta;
                  real8 detC23 = dx_xi   * dz_eta  - dz_xi   * dx_eta;
                  real8 detC33 = dx_xi   * dy_eta  - dy_xi   + dx_eta;

                  // -------------------

                  real8 b11 = detC11/detMt;
                  real8 b21 = detC21/detMt;
                  real8 b31 = detC31/detMt;

                  real8 b12 = detC12/detMt;
                  real8 b22 = detC22/detMt;
                  real8 b32 = detC32/detMt;

                  real8 b13 = detC13/detMt;
                  real8 b23 = detC23/detMt;
                  real8 b33 = detC33/detMt;

                  //
                  // determine the maximum gradient in x, y and z (nice orthonormal basis)
                  //
                  dv_x[0] = b11*dv_xi[0] + b12*dv_xi[1] + b13*dv_xi[2];
                  dv_x[1] = b21*dv_xi[0] + b22*dv_xi[1] + b23*dv_xi[2];
                  dv_x[2] = b31*dv_xi[0] + b32*dv_xi[1] + b33*dv_xi[2];

                  double vmax = MAX( dv_x[0], MAX( dv_x[1], dv_x[2] ) );

                  if ( vmax > tol )
                     ltags[tind] = TRUE;
               }
            }
         }

      } // criteria = UVAL_GRADIENT

      if (ref == "USER_DEFINED") {

         /*
          * For user-defined, access refine box data from the MblkGeometry
          * class.
          */
         int level_number = patch.getPatchLevelNumber();
         int block_number = getBlockNumber();
         hier::BoxArray<NDIM> refine_boxes(1);
         if (d_mblk_geometry->getRefineBoxes(refine_boxes,
                                             block_number,
                                             level_number) ) {
            for (int b = 0; b < refine_boxes.getNumberOfBoxes(); b++) {
               hier::Box<NDIM> intersect = pbox * refine_boxes[b];
               if (!intersect.empty()) {
                  temp_tags->fill(TRUE,intersect);
               }
            }
         }

      } // criteria = USER_DEFINED

   }  // loop over criteria

   //
   // Update tags
   //
   for (pdat::CellIterator<NDIM> ic(pbox ); ic; ic++) {
      (*tags)(ic(),0) = (*temp_tags)(ic(),0 );
   }

   tbox::plog << "--------------------- end tagGradientCells" << endl;

}


/*
*************************************************************************
*                                                                       *
* Fill the singularity conditions for the multi-block case
*                                                                       *
*************************************************************************
*/
void MblkLinAdv::fillSingularityBoundaryConditions(
   hier::Patch<NDIM>& patch,
   tbox::List<xfer::MultiblockRefineSchedule<NDIM>::SingularityPatch>&
      sing_patches,
   const double fill_time,
   const hier::Box<NDIM>& fill_box,
   const hier::BoundaryBox<NDIM>& bbox)
{

   NULL_USE(patch);
   NULL_USE(sing_patches);
   NULL_USE(fill_time);
   NULL_USE(fill_box);
   NULL_USE(bbox);

   /*
    * Here we need to specify how to fill the singularities...
    */
#if 0
   for (int i = 0; i < d_variables.getSize(); i++) {

      tbox::Pointer< pdat::CellData<NDIM,double> > cell_data =
         patch.getPatchData(d_variables[i], getDataContext());

      hier::Box<NDIM> sing_fill_box(cell_data->getGhostBox() * fill_box);
      cell_data->fillAll(0.0, sing_fill_box);

      int depth = cell_data->getDepth();

      /*
       * If sing_patches is not empty, that means there is enhanced
       * connectivity, and we get data from other blocks
       */

      if (sing_patches.size()) {

         for (tbox::List
              <xfer::MultiblockRefineSchedule<NDIM>::SingularityPatch>::
              Iterator sp(sing_patches); sp; sp++) {
            tbox::Pointer< pdat::CellData<NDIM,double> > sing_data =
               sp().d_patch->getPatchData(d_variables[i], getDataContext());
            int sing_neighbor_id = sp().d_id;
            for (pdat::CellIterator<NDIM> ci(sing_fill_box); ci; ci++) {
               for (int d = 0; d < depth; d++) {
                  (*cell_data)(ci(),d) += sing_neighbor_id;
               }
            }
         }

         for (pdat::CellIterator<NDIM> ci(sing_fill_box); ci; ci++) {
            for (int d = 0; d < depth; d++) {
               (*cell_data)(ci(),d) /= sing_patches.size();
            }
         }

      /*
       * In cases of reduced connectivity, there are no other blocks
       * from which to acquire data.
       */

      } else {

         cell_data->fillAll(
            (double)bbox.getLocationIndex()+200.0, fill_box);

      }
   }
#endif

}

/*
*************************************************************************
*                                                                       *
* Private method to build XYZ coordinates on a patch                    *
*                                                                       *
*************************************************************************
*/
void MblkLinAdv::setMappedGridOnPatch(const hier::Patch<NDIM>& patch,
                                      const int level_number,
                                      const int block_number)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   TBOX_ASSERT(level_number >= 0);
#endif

   // compute level domain
   const tbox::Pointer<geom::BlockPatchGeometry<NDIM> >
      patch_geom = patch.getPatchGeometry();
   hier::IntVector<NDIM> ratio = patch_geom->getRatio();
   hier::BoxArray<NDIM> domain_boxes;
   d_grid_geometries[block_number]->computePhysicalDomain(domain_boxes, ratio);
   int num_domain_boxes = domain_boxes.getNumberOfBoxes();

   if (num_domain_boxes > 1) {
      TBOX_ERROR("Sorry, cannot handle non-rectangular domains..." << endl);
   }

   int xyz_id =  hier::VariableDatabase<NDIM>::getDatabase()->
      mapVariableAndContextToIndex(d_xyz, getDataContext());

   d_mblk_geometry->buildGridOnPatch(patch,
                                     domain_boxes[0],
                                     xyz_id,
                                     level_number,
                                     block_number);
}




/*
*************************************************************************
*                                                                       *
* Register VisIt data writer to write data to plot files that may       *
* be postprocessed by the VisIt tool.                                   *
*                                                                       *
*************************************************************************
*/

#ifdef HAVE_HDF5
void MblkLinAdv::registerVisItDataWriter(
   tbox::Pointer<appu::VisItDataWriter<NDIM> > viz_writer)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   TBOX_ASSERT(!(viz_writer.isNull()));
#endif
   d_visit_writer = viz_writer;
}
#endif

/*
*************************************************************************
*                                                                       *
* Write MblkLinAdv object state to specified stream.                        *
*                                                                       *
*************************************************************************
*/

void MblkLinAdv::printClassData(ostream &os) const
{
   int j,k;

   os << "\nMblkLinAdv::printClassData..." << endl;
   os << "MblkLinAdv: this = " << (MblkLinAdv*)this << endl;
   os << "d_object_name = " << d_object_name << endl;
   os << "d_grid_geometries = " << endl;
//   for (j=0; j < d_grid_geometries.getSize(); j++) {
//      os << (*((tbox::Pointer<geom::BlockGridGeometry<NDIM> >)(d_grid_geometries[j]))) << endl;
//   }

   os << "Parameters for numerical method ..." << endl;
   os << "   d_advection_velocity = ";
   for (j=0;j<NDIM;j++) os << d_advection_velocity[j] << " ";
   os << endl;
   os << "   d_godunov_order = " << d_godunov_order << endl;
   os << "   d_corner_transport = " << d_corner_transport << endl;
   os << "   d_nghosts = " << d_nghosts << endl;
   os << "   d_fluxghosts = " << d_fluxghosts << endl;

   os << "Problem description and initial data..." << endl;
   os << "   d_data_problem = " << d_data_problem << endl;
   os << "   d_data_problem_int = " << d_data_problem << endl;

   os << "       d_radius = " << d_radius << endl;
   os << "       d_center = " ;
      for (j=0;j<NDIM;j++) os << d_center[j]<<" ";
   os << endl;
   os << "       d_uval_inside = " << d_uval_inside << endl;
   os << "       d_uval_outside = " << d_uval_outside << endl;

   os << "       d_number_of_intervals = " << d_number_of_intervals << endl;
   os << "       d_front_position = ";
   for (k = 0; k < d_number_of_intervals-1; k++) {
      os << d_front_position[k] << "  ";
   }
   os << endl;
   os << "       d_interval_uval = " << endl;
   for (k = 0; k < d_number_of_intervals; k++) {
      os << "            " << d_interval_uval[k] << endl;
   }
   os << "   Boundary condition data " << endl;

#if (NDIM == 2)
   for (j = 0; j < d_scalar_bdry_edge_conds.getSize(); j++) {
      os << "       d_scalar_bdry_edge_conds[" << j << "] = "
         << d_scalar_bdry_edge_conds[j] << endl;
      if (d_scalar_bdry_edge_conds[j] == DIRICHLET_BC) {
         os << "         d_bdry_edge_uval[" << j << "] = "
            << d_bdry_edge_uval[j] << endl;
      }
   }
   os << endl;
   for (j = 0; j < d_scalar_bdry_node_conds.getSize(); j++) {
      os << "       d_scalar_bdry_node_conds[" << j << "] = "
         << d_scalar_bdry_node_conds[j] << endl;
      os << "       d_node_bdry_edge[" << j << "] = "
         << d_node_bdry_edge[j] << endl;
   }
#endif
#if (NDIM == 3)
   for (j = 0; j < d_scalar_bdry_face_conds.getSize(); j++) {
      os << "       d_scalar_bdry_face_conds[" << j << "] = "
         << d_scalar_bdry_face_conds[j] << endl;
      if (d_scalar_bdry_face_conds[j] == DIRICHLET_BC) {
         os << "         d_bdry_face_uval[" << j << "] = "
            << d_bdry_face_uval[j] << endl;
      }
   }
   os << endl;
   for (j = 0; j < d_scalar_bdry_edge_conds.getSize(); j++) {
      os << "       d_scalar_bdry_edge_conds[" << j << "] = "
         << d_scalar_bdry_edge_conds[j] << endl;
      os << "       d_edge_bdry_face[" << j << "] = "
         << d_edge_bdry_face[j] << endl;
   }
   os << endl;
   for (j = 0; j < d_scalar_bdry_node_conds.getSize(); j++) {
      os << "       d_scalar_bdry_node_conds[" << j << "] = "
         << d_scalar_bdry_node_conds[j] << endl;
      os << "       d_node_bdry_face[" << j << "] = "
         << d_node_bdry_face[j] << endl;
   }
#endif

   os << "   Refinement criteria parameters " << endl;

   for (j = 0; j < d_refinement_criteria.getSize(); j++) {
      os << "       d_refinement_criteria[" << j << "] = "
         << d_refinement_criteria[j] << endl;
   }
   os << endl;
   for (j = 0; j < d_dev_tol.getSize(); j++) {
      os << "       d_dev_tol[" << j << "] = "
         << d_dev_tol[j] << endl;
   }
  for (j = 0; j < d_dev.getSize(); j++) {
      os << "       d_dev[" << j << "] = "
         << d_dev[j] << endl;
   }
   os << endl;
   for (j = 0; j < d_dev_time_max.getSize(); j++) {
      os << "       d_dev_time_max[" << j << "] = "
         << d_dev_time_max[j] << endl;
   }
   os << endl;
   for (j = 0; j < d_dev_time_min.getSize(); j++) {
      os << "       d_dev_time_min[" << j << "] = "
         << d_dev_time_min[j] << endl;
   }
   os << endl;
   for (j = 0; j < d_grad_tol.getSize(); j++) {
      os << "       d_grad_tol[" << j << "] = "
         << d_grad_tol[j] << endl;
   }
   os << endl;
   for (j = 0; j < d_grad_time_max.getSize(); j++) {
      os << "       d_grad_time_max[" << j << "] = "
         << d_grad_time_max[j] << endl;
   }
   os << endl;
   for (j = 0; j < d_grad_time_min.getSize(); j++) {
      os << "       d_grad_time_min[" << j << "] = "
         << d_grad_time_min[j] << endl;
   }
   os << endl;
   for (j = 0; j < d_shock_onset.getSize(); j++) {
      os << "       d_shock_onset[" << j << "] = "
         << d_shock_onset[j] << endl;
   }
   os << endl;
   for (j = 0; j < d_shock_tol.getSize(); j++) {
      os << "       d_shock_tol[" << j << "] = "
         << d_shock_tol[j] << endl;
   }
   os << endl;
   for (j = 0; j < d_shock_time_max.getSize(); j++) {
      os << "       d_shock_time_max[" << j << "] = "
         << d_shock_time_max[j] << endl;
   }
   os << endl;
   for (j = 0; j < d_shock_time_min.getSize(); j++) {
      os << "       d_shock_time_min[" << j << "] = "
         << d_shock_time_min[j] << endl;
   }
   os << endl;
   for (j = 0; j < d_rich_tol.getSize(); j++) {
      os << "       d_rich_tol[" << j << "] = "
         << d_rich_tol[j] << endl;
   }
   os << endl;
   for (j = 0; j < d_rich_time_max.getSize(); j++) {
      os << "       d_rich_time_max[" << j << "] = "
         << d_rich_time_max[j] << endl;
   }
   os << endl;
   for (j = 0; j < d_rich_time_min.getSize(); j++) {
      os << "       d_rich_time_min[" << j << "] = "
         << d_rich_time_min[j] << endl;
   }
   os << endl;

}


/*
*************************************************************************
*                                                                       *
* Read data members from input.  All values set from restart can be	*
* overridden by values in the input database.
*                                                                       *
*************************************************************************
*/
void MblkLinAdv::getFromInput(
   tbox::Pointer<tbox::Database> input_db,
   bool is_from_restart)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   TBOX_ASSERT(!input_db.isNull());
#endif

   tbox::Pointer<tbox::Database> db = input_db->getDatabase("MblkLinAdv");

   /*
    * Note: if we are restarting, then we only allow nonuniform
    * workload to be used if nonuniform workload was used originally.
    */
   if (!is_from_restart) {
      d_use_nonuniform_workload =
         db->getBoolWithDefault("use_nonuniform_workload",
                                d_use_nonuniform_workload);
   } else {
      if (d_use_nonuniform_workload) {
         d_use_nonuniform_workload =
            db->getBool("use_nonuniform_workload");
      }
   }

   if (db->keyExists("advection_velocity")){
      db->getDoubleArray("advection_velocity",
                         d_advection_velocity, NDIM);
   } else {
      TBOX_ERROR(d_object_name << ":  "
         << "Key data `advection_velocity' not found in input.");
   }

   if (db->keyExists("godunov_order")) {
      d_godunov_order = db->getInteger("godunov_order");
      if ( (d_godunov_order != 1) &&
           (d_godunov_order != 2) &&
           (d_godunov_order != 4) ) {
         TBOX_ERROR(d_object_name << ": "
                    << "`godunov_order' in input must be 1, 2, or 4." << endl);
      }
   } else {
      d_godunov_order = db->getIntegerWithDefault("d_godunov_order",
                                                   d_godunov_order);
   }

   if (db->keyExists("corner_transport")) {
      d_corner_transport = db->getString("corner_transport");
      if ( (d_corner_transport != "CORNER_TRANSPORT_1") &&
           (d_corner_transport != "CORNER_TRANSPORT_2") ) {
         TBOX_ERROR(d_object_name << ": "
            << "`corner_transport' in input must be either string"
            << " 'CORNER_TRANSPORT_1' or 'CORNER_TRANSPORT_2'." << endl);
      }
   } else {
      d_corner_transport = db->getStringWithDefault("corner_transport",
                                                     d_corner_transport);
   }

   if (db->keyExists("Refinement_data")){
      tbox::Pointer<tbox::Database> refine_db = db->getDatabase("Refinement_data");
      tbox::Array<string> refinement_keys = refine_db->getAllKeys();
      int num_keys = refinement_keys.getSize();

      if (refine_db->keyExists("refine_criteria")) {
         d_refinement_criteria =
           refine_db->getStringArray("refine_criteria");
      } else {
        TBOX_WARNING(d_object_name << ": "
                   << "No key `refine_criteria' found in data for"
                   << " RefinementData. No refinement will occur." << endl);
      }

      tbox::Array<string> ref_keys_defined(num_keys);
      int def_key_cnt = 0;
      tbox::Pointer<tbox::Database> error_db;
      for (int i = 0; i < refinement_keys.getSize(); i++) {

         string error_key = refinement_keys[i];
         error_db.setNull();

         if ( !(error_key == "refine_criteria") ) {

            if ( !(error_key == "UVAL_DEVIATION" ||
                   error_key == "UVAL_GRADIENT" ||
                   error_key == "UVAL_SHOCK" ||
                   error_key == "UVAL_RICHARDSON") ) {
               TBOX_ERROR(d_object_name << ": "
                          << "Unknown refinement criteria: " << error_key
                          << "\nin input." << endl);
            } else {
               error_db = refine_db->getDatabase(error_key);
               ref_keys_defined[def_key_cnt] = error_key;
               def_key_cnt++;
            }

            if (!error_db.isNull() && error_key == "UVAL_DEVIATION") {

               if (error_db->keyExists("dev_tol")) {
                  d_dev_tol =
                     error_db->getDoubleArray("dev_tol");
               } else {
                  TBOX_ERROR(d_object_name << ": "
                             << "No key `dev_tol' found in data for "
                             << error_key << endl);
               }

               if (error_db->keyExists("uval_dev")) {
                  d_dev =
                     error_db->getDoubleArray("uval_dev");
               } else {
                  TBOX_ERROR(d_object_name << ": "
                             << "No key `uval_dev' found in data for "
                             << error_key << endl);
               }

               if (error_db->keyExists("time_max")){
                  d_dev_time_max =
                     error_db->getDoubleArray("time_max");
               } else {
                  d_dev_time_max.resizeArray(1);
                  d_dev_time_max[0] = tbox::MathUtilities<double>::getMax();
               }

               if (error_db->keyExists("time_min")){
                  d_dev_time_min =
                     error_db->getDoubleArray("time_min");
               } else {
                  d_dev_time_min.resizeArray(1);
                  d_dev_time_min[0] = 0.;
               }

            }

            if (!error_db.isNull() && error_key == "UVAL_GRADIENT") {

               if (error_db->keyExists("grad_tol")) {
                  d_grad_tol =
                     error_db->getDoubleArray("grad_tol");
               } else {
                  TBOX_ERROR(d_object_name << ": "
                             << "No key `grad_tol' found in data for "
                             << error_key << endl);
               }

               if (error_db->keyExists("time_max")){
                  d_grad_time_max =
                     error_db->getDoubleArray("time_max");
               } else {
                  d_grad_time_max.resizeArray(1);
                  d_grad_time_max[0] = tbox::MathUtilities<double>::getMax();
               }

               if (error_db->keyExists("time_min")){
                  d_grad_time_min =
                     error_db->getDoubleArray("time_min");
               } else {
                  d_grad_time_min.resizeArray(1);
                  d_grad_time_min[0] = 0.;
               }

            }

            if (!error_db.isNull() && error_key == "UVAL_SHOCK") {

               if (error_db->keyExists("shock_onset")) {
                  d_shock_onset =
                     error_db->getDoubleArray("shock_onset");
               } else {
                  TBOX_ERROR(d_object_name << ": "
                             << "No key `shock_onset' found in data for "
                             << error_key << endl);
               }

               if (error_db->keyExists("shock_tol")) {
                  d_shock_tol =
                     error_db->getDoubleArray("shock_tol");
               } else {
                  TBOX_ERROR(d_object_name << ": "
                             << "No key `shock_tol' found in data for "
                             << error_key << endl);
               }

               if (error_db->keyExists("time_max")){
                  d_shock_time_max =
                     error_db->getDoubleArray("time_max");
               } else {
                  d_shock_time_max.resizeArray(1);
                  d_shock_time_max[0] = tbox::MathUtilities<double>::getMax();
               }

               if (error_db->keyExists("time_min")){
                  d_shock_time_min =
                     error_db->getDoubleArray("time_min");
               } else {
                  d_shock_time_min.resizeArray(1);
                  d_shock_time_min[0] = 0.;
               }

            }

            if (!error_db.isNull() && error_key == "UVAL_RICHARDSON") {

               if (error_db->keyExists("rich_tol")) {
                  d_rich_tol =
                     error_db->getDoubleArray("rich_tol");
               } else {
                  TBOX_ERROR(d_object_name << ": "
                             << "No key `rich_tol' found in data for "
                             << error_key << endl);
               }

               if (error_db->keyExists("time_max")){
                  d_rich_time_max =
                     error_db->getDoubleArray("time_max");
               } else {
                  d_rich_time_max.resizeArray(1);
                  d_rich_time_max[0] = tbox::MathUtilities<double>::getMax();
               }

               if (error_db->keyExists("time_min")){
                  d_rich_time_min =
                     error_db->getDoubleArray("time_min");
               } else {
                  d_rich_time_min.resizeArray(1);
                  d_rich_time_min[0] = 0.;
               }

            }

         }

      } // loop over refine criteria

   } // refine db entry exists

   if (!is_from_restart) {

      if (db->keyExists("data_problem")) {
         d_data_problem = db->getString("data_problem");
      } else {
         TBOX_ERROR(d_object_name << ": "
                  << "`data_problem' value not found in input." << endl);
      }

      tbox::Pointer<tbox::Database> init_data_db;
      if (db->keyExists("Initial_data")) {
         init_data_db = db->getDatabase("Initial_data");
      } else {
         TBOX_ERROR(d_object_name << ": "
                  << "No `Initial_data' database found in input." << endl);
      }

      bool found_problem_data = false;

      if (d_data_problem == "SPHERE") {

         if (init_data_db->keyExists("radius")) {
            d_radius = init_data_db->getDouble("radius");
         } else {
            TBOX_ERROR(d_object_name << ": "
               << "`radius' input required for SPHERE problem." << endl);
         }
         if (init_data_db->keyExists("center")) {
            init_data_db->getDoubleArray("center", d_center, NDIM);
         } else {
            TBOX_ERROR(d_object_name << ": "
               << "`center' input required for SPHERE problem." << endl);
         }
         if (init_data_db->keyExists("uval_inside")) {
            d_uval_inside = init_data_db->getDouble("uval_inside");
         } else {
            TBOX_ERROR(d_object_name << ": "
               << "`uval_inside' input required for "
               << "SPHERE problem." << endl);
         }
         if (init_data_db->keyExists("uval_outside")) {
            d_uval_outside = init_data_db->getDouble("uval_outside");
         } else {
            TBOX_ERROR(d_object_name << ": "
               << "`uval_outside' input required for "
               << "SPHERE problem." << endl);
         }

         found_problem_data = true;

      }

      if (!found_problem_data &&
          (d_data_problem == "PIECEWISE_CONSTANT_X") ||
          (d_data_problem == "PIECEWISE_CONSTANT_Y") ||
          (d_data_problem == "PIECEWISE_CONSTANT_Z") ||
          (d_data_problem == "SINE_CONSTANT_X") ||
          (d_data_problem == "SINE_CONSTANT_Y") ||
          (d_data_problem == "SINE_CONSTANT_Z")) {

          int idir = 0;
          if (d_data_problem == "PIECEWISE_CONSTANT_Y") {
             if (NDIM < 2) {
                TBOX_ERROR(d_object_name << ": `PIECEWISE_CONSTANT_Y' "
                           << "problem invalid in 1 dimension." << endl);
             }
             idir = 1;
          }

          if (d_data_problem == "PIECEWISE_CONSTANT_Z") {
             if (NDIM < 3) {
                TBOX_ERROR(d_object_name << ": `PIECEWISE_CONSTANT_Z' "
                           << "problem invalid in 1 or 2 dimensions." << endl);
             }
             idir = 2;
          }

          tbox::Array<string> init_data_keys = init_data_db->getAllKeys();

          if (init_data_db->keyExists("front_position")) {
             d_front_position = init_data_db->getDoubleArray("front_position");
          } else {
             TBOX_ERROR(d_object_name << ": "
                << "`front_position' input required for "
                << d_data_problem << " problem." << endl);
          }

          d_number_of_intervals =
             tbox::MathUtilities<int>::Min( d_front_position.getSize()+1,
                                            init_data_keys.getSize()-1 );

          d_interval_uval.resizeArray(d_number_of_intervals);

          int i = 0;
          int nkey = 0;
          bool found_interval_data = false;

          while (   !found_interval_data
                 && (i < d_number_of_intervals)
                 && (nkey < init_data_keys.getSize()) ) {

             if ( !(init_data_keys[nkey] == "front_position") ) {

                tbox::Pointer<tbox::Database> interval_db =
                   init_data_db->getDatabase(init_data_keys[nkey]);

                if (interval_db->keyExists("uval")) {
                   d_interval_uval[i] = interval_db->getDouble("uval");
                } else {
                   TBOX_ERROR(d_object_name << ": "
                          << "`uval' data missing in input for key = "
                          << init_data_keys[nkey] << endl);
                }
                i++;

                found_interval_data = (i == d_number_of_intervals);

             }

             nkey++;

          }

          if ((d_data_problem == "SINE_CONSTANT_X") ||
              (d_data_problem == "SINE_CONSTANT_Y") ||
              (d_data_problem == "SINE_CONSTANT_Z")) {
               if (init_data_db->keyExists("amplitude")) {
                  d_amplitude = init_data_db->getDouble("amplitude");
               }
               if (init_data_db->keyExists("frequency")) {
                  init_data_db->getDoubleArray("frequency",d_frequency,NDIM);
               } else {
                  TBOX_ERROR(d_object_name << ": "
                   << "`frequency' input required for SINE problem." << endl);
               }
          }

          if (!found_interval_data) {
             TBOX_ERROR(d_object_name << ": "
                        << "Insufficient interval data given in input"
                        << " for PIECEWISE_CONSTANT_*problem." << endl);
          }

          found_problem_data = true;
      }

      if (!found_problem_data) {
         TBOX_ERROR(d_object_name << ": "
            << "`Initial_data' database found in input."
            << " But bad data supplied." << endl);
      }

   } // if !is_from_restart read in problem data

   hier::IntVector<NDIM> periodic = d_grid_geometries[0]->getPeriodicShift();
   int num_per_dirs = 0;
   for (int id = 0; id < NDIM; id++) {
      if (periodic(id)) num_per_dirs++;
   }

   /*
    * If there are multiple blocks, periodicity is not currently supported.
    */
   if ((d_grid_geometries.getSize() > 1) && (num_per_dirs > 0)) {
      TBOX_ERROR(d_object_name << ": cannot have periodic BCs when there"
                 << "\nare multiple blocks." << endl);
   }


   if (db->keyExists("Boundary_data")) {

      tbox::Pointer<tbox::Database> bdry_db = db->getDatabase("Boundary_data");

#if (NDIM == 2)
      SkeletonBoundaryUtilities2::readBoundaryInput(this,
                                                     bdry_db,
                                                     d_scalar_bdry_edge_conds,
                                                     d_scalar_bdry_node_conds,
                                                     periodic);
#endif
#if (NDIM == 3)
      SkeletonBoundaryUtilities3::readBoundaryInput(this,
                                                     bdry_db,
                                                     d_scalar_bdry_face_conds,
                                                     d_scalar_bdry_edge_conds,
                                                     d_scalar_bdry_node_conds,
                                                     periodic);
#endif

   } else {
      TBOX_ERROR(d_object_name << ": "
                 << "Key data `Boundary_data' not found in input. " << endl);
   }


}


/*
*************************************************************************
*                                                                       *
* Routines to put/get data members to/from restart database.            *
*                                                                       *
*************************************************************************
*/

void MblkLinAdv::putToDatabase(
   tbox::Pointer<tbox::Database> db)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   TBOX_ASSERT(!db.isNull());
#endif

   db->putInteger("MBLKLINADV_VERSION",MBLKLINADV_VERSION);

   db->putDoubleArray("d_advection_velocity",d_advection_velocity,NDIM);

   db->putInteger("d_godunov_order", d_godunov_order);
   db->putString("d_corner_transport", d_corner_transport);
   db->putIntegerArray("d_nghosts", (int*)d_nghosts, NDIM);
   db->putIntegerArray("d_fluxghosts", (int*)d_fluxghosts, NDIM);

   db->putString("d_data_problem", d_data_problem);

   if (d_data_problem == "SPHERE") {
      db->putDouble("d_radius", d_radius);
      db->putDoubleArray("d_center", d_center, NDIM);
      db->putDouble("d_uval_inside", d_uval_inside);
      db->putDouble("d_uval_outside", d_uval_outside);
   }

   if ( (d_data_problem == "PIECEWISE_CONSTANT_X") ||
        (d_data_problem == "PIECEWISE_CONSTANT_Y") ||
        (d_data_problem == "PIECEWISE_CONSTANT_Z") ||
        (d_data_problem == "SINE_CONSTANT_X") ||
        (d_data_problem == "SINE_CONSTANT_Y") ||
        (d_data_problem == "SINE_CONSTANT_Z") ) {
      db->putInteger("d_number_of_intervals", d_number_of_intervals);
      if (d_number_of_intervals > 0) {
         db->putDoubleArray("d_front_position", d_front_position);
         db->putDoubleArray("d_interval_uval", d_interval_uval);
      }
   }

   db->putIntegerArray("d_scalar_bdry_edge_conds", d_scalar_bdry_edge_conds);
   db->putIntegerArray("d_scalar_bdry_node_conds", d_scalar_bdry_node_conds);

#if (NDIM == 2)
   db->putDoubleArray("d_bdry_edge_uval", d_bdry_edge_uval);
#endif
#if (NDIM == 3)
   db->putIntegerArray("d_scalar_bdry_face_conds", d_scalar_bdry_face_conds);
   db->putDoubleArray("d_bdry_face_uval", d_bdry_face_uval);
#endif

   if (d_refinement_criteria.getSize() > 0) {
      db->putStringArray("d_refinement_criteria", d_refinement_criteria);
   }
   for (int i = 0; i < d_refinement_criteria.getSize(); i++) {

      if (d_refinement_criteria[i] == "UVAL_DEVIATION") {
         db->putDoubleArray("d_dev_tol", d_dev_tol);
         db->putDoubleArray("d_dev", d_dev);
         db->putDoubleArray("d_dev_time_max", d_dev_time_max);
         db->putDoubleArray("d_dev_time_min", d_dev_time_min);
      } else if (d_refinement_criteria[i] == "UVAL_GRADIENT") {
         db->putDoubleArray("d_grad_tol", d_grad_tol);
         db->putDoubleArray("d_grad_time_max", d_grad_time_max);
         db->putDoubleArray("d_grad_time_min", d_grad_time_min);
      } else if  (d_refinement_criteria[i] == "UVAL_SHOCK") {
         db->putDoubleArray("d_shock_onset", d_shock_onset);
         db->putDoubleArray("d_shock_tol", d_shock_tol);
         db->putDoubleArray("d_shock_time_max", d_shock_time_max);
         db->putDoubleArray("d_shock_time_min", d_shock_time_min);
      } else if  (d_refinement_criteria[i] == "UVAL_RICHARDSON") {
         db->putDoubleArray("d_rich_tol", d_rich_tol);
         db->putDoubleArray("d_rich_time_max", d_rich_time_max);
         db->putDoubleArray("d_rich_time_min", d_rich_time_min);
      }

   }

}


/*
*************************************************************************
*                                                                       *
*    Access class information from restart database.                    *
*                                                                       *
*************************************************************************
*/
void MblkLinAdv::getFromRestart()
{
   tbox::Pointer<tbox::Database> root_db =
      tbox::RestartManager::getManager()->getRootDatabase();

   tbox::Pointer<tbox::Database> db;
   if ( root_db->isDatabase(d_object_name) ) {
      db = root_db->getDatabase(d_object_name);
   } else {
      TBOX_ERROR("Restart database corresponding to "
              << d_object_name << " not found in restart file.");
   }

   int ver = db->getInteger("MBLKLINADV_VERSION");
   if (ver != MBLKLINADV_VERSION){
      TBOX_ERROR(d_object_name << ":  "
              << "Restart file version different than class version.");
   }

   db->getDoubleArray("d_advection_velocity",d_advection_velocity,NDIM);

   d_godunov_order = db->getInteger("d_godunov_order");
   d_corner_transport = db->getString("d_corner_transport");

#if 0
   int* tmp_nghosts = d_nghosts;
   db->getIntegerArray("d_nghosts", tmp_nghosts, NDIM);
   if ( !(d_nghosts == CELLG) ) {
      TBOX_ERROR(d_object_name << ": "
                 << "Key data `d_nghosts' in restart file != CELLG." << endl);
   }
   int* tmp_fluxghosts = d_fluxghosts;
   db->getIntegerArray("d_fluxghosts", tmp_fluxghosts, NDIM);
   if ( !(d_fluxghosts == FLUXG) ) {
      TBOX_ERROR(d_object_name << ": "
         << "Key data `d_fluxghosts' in restart file != FLUXG." << endl);
   }
#endif

   d_data_problem = db->getString("d_data_problem");

   if (d_data_problem == "SPHERE") {
      d_data_problem_int = SPHERE;
      d_radius = db->getDouble("d_radius");
      db->getDoubleArray("d_center", d_center, NDIM);
      d_uval_inside = db->getDouble("d_uval_inside");
      d_uval_outside = db->getDouble("d_uval_outside");
   }

   if ( (d_data_problem == "PIECEWISE_CONSTANT_X") ||
        (d_data_problem == "PIECEWISE_CONSTANT_Y") ||
        (d_data_problem == "PIECEWISE_CONSTANT_Z") ||
        (d_data_problem == "SINE_CONSTANT_X") ||
        (d_data_problem == "SINE_CONSTANT_Y") ||
        (d_data_problem == "SINE_CONSTANT_Z") ) {
      d_number_of_intervals = db->getInteger("d_number_of_intervals");
      if (d_number_of_intervals > 0) {
         d_front_position = db->getDoubleArray("d_front_position");
         d_interval_uval = db->getDoubleArray("d_interval_uval");
      }
   }

   d_scalar_bdry_edge_conds = db->getIntegerArray("d_scalar_bdry_edge_conds");
   d_scalar_bdry_node_conds = db->getIntegerArray("d_scalar_bdry_node_conds");

#if (NDIM == 2)
   d_bdry_edge_uval = db->getDoubleArray("d_bdry_edge_uval");
#endif
#if (NDIM == 3)
   d_scalar_bdry_face_conds = db->getIntegerArray("d_scalar_bdry_face_conds");

   d_bdry_face_uval = db->getDoubleArray("d_bdry_face_uval");
#endif

   if (db->keyExists("d_refinement_criteria")) {
      d_refinement_criteria = db->getStringArray("d_refinement_criteria");
   }
   for (int i = 0; i < d_refinement_criteria.getSize(); i++) {

      if (d_refinement_criteria[i] == "UVAL_DEVIATION") {
         d_dev_tol = db->getDoubleArray("d_dev_tol");
         d_dev_time_max = db->getDoubleArray("d_dev_time_max");
         d_dev_time_min = db->getDoubleArray("d_dev_time_min");
      } else if (d_refinement_criteria[i] == "UVAL_GRADIENT") {
         d_grad_tol = db->getDoubleArray("d_grad_tol");
         d_grad_time_max = db->getDoubleArray("d_grad_time_max");
         d_grad_time_min = db->getDoubleArray("d_grad_time_min");
      } else if  (d_refinement_criteria[i] == "UVAL_SHOCK") {
         d_shock_onset = db->getDoubleArray("d_shock_onset");
         d_shock_tol = db->getDoubleArray("d_shock_tol");
         d_shock_time_max = db->getDoubleArray("d_shock_time_max");
         d_shock_time_min = db->getDoubleArray("d_shock_time_min");
      } else if  (d_refinement_criteria[i] == "UVAL_RICHARDSON") {
         d_rich_tol = db->getDoubleArray("d_rich_tol");
         d_rich_time_max = db->getDoubleArray("d_rich_time_max");
         d_rich_time_min = db->getDoubleArray("d_rich_time_min");
      }

   }

}

/*
*************************************************************************
*                                                                       *
* Routines to read boundary data from input database.                   *
*                                                                       *
*************************************************************************
*/

void MblkLinAdv::readDirichletBoundaryDataEntry(tbox::Pointer<tbox::Database> db,
                                           string& db_name,
                                           int bdry_location_index)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   TBOX_ASSERT(!db.isNull());
   TBOX_ASSERT(!db_name.empty());
#endif
#if (NDIM == 2)
   readStateDataEntry(db,
                      db_name,
                      bdry_location_index,
                      d_bdry_edge_uval);
#endif
#if (NDIM == 3)
   readStateDataEntry(db,
                      db_name,
                      bdry_location_index,
                      d_bdry_face_uval);
#endif
}

void MblkLinAdv::readStateDataEntry(tbox::Pointer<tbox::Database> db,
                               const string& db_name,
                               int array_indx,
                               tbox::Array<double>& uval)
{
#ifdef DEBUG_CHECK_ASSERTIONS
   TBOX_ASSERT(!db.isNull());
   TBOX_ASSERT(!db_name.empty());
   TBOX_ASSERT(array_indx >= 0);
   TBOX_ASSERT(uval.getSize() > array_indx);
#endif

   if (db->keyExists("uval")) {
      uval[array_indx] = db->getDouble("uval");
   } else {
      TBOX_ERROR(d_object_name << ": "
         << "`uval' entry missing from " << db_name
         << " input database. " << endl);
   }

}

/*
*************************************************************************
*                                                                       *
* Routine to check boundary data when debugging.                        *
*                                                                       *
*************************************************************************
*/

void MblkLinAdv::checkBoundaryData(int btype,
                              const hier::Patch<NDIM>& patch,
                              const hier::IntVector<NDIM>& ghost_width_to_check,
                              const tbox::Array<int>& scalar_bconds) const
{
#ifdef DEBUG_CHECK_ASSERTIONS
#if (NDIM == 2)
   TBOX_ASSERT(btype == EDGE2D_BDRY_TYPE ||
          btype == NODE2D_BDRY_TYPE);
#endif
#if (NDIM == 3)
   TBOX_ASSERT(btype == FACE3D_BDRY_TYPE ||
          btype == EDGE3D_BDRY_TYPE ||
          btype == NODE3D_BDRY_TYPE);
#endif
#endif

   const tbox::Pointer<geom::BlockPatchGeometry<NDIM> > pgeom = patch.getPatchGeometry();
   const tbox::Array<hier::BoundaryBox<NDIM> > bdry_boxes =
      pgeom->getCodimensionBoundaries(btype);

   hier::VariableDatabase<NDIM>* vdb = hier::VariableDatabase<NDIM>::getDatabase();

   for (int i = 0; i < bdry_boxes.getSize(); i++ ) {
      hier::BoundaryBox<NDIM> bbox = bdry_boxes[i];
#ifdef DEBUG_CHECK_ASSERTIONS
      TBOX_ASSERT(bbox.getBoundaryType() == btype);
#endif
      int bloc = bbox.getLocationIndex();

      int bscalarcase, refbdryloc;
#if (NDIM == 2)
      if (btype == EDGE2D_BDRY_TYPE) {
#ifdef DEBUG_CHECK_ASSERTIONS
         TBOX_ASSERT(scalar_bconds.getSize() == NUM_2D_EDGES);
#endif
         bscalarcase = scalar_bconds[bloc];
         refbdryloc = bloc;
      } else { // btype == NODE2D_BDRY_TYPE
#ifdef DEBUG_CHECK_ASSERTIONS
         TBOX_ASSERT(scalar_bconds.getSize() == NUM_2D_NODES);
#endif
         bscalarcase = scalar_bconds[bloc];
         refbdryloc = d_node_bdry_edge[bloc];
      }
#endif
#if (NDIM == 3)
      if (btype == FACE3D_BDRY_TYPE) {
#ifdef DEBUG_CHECK_ASSERTIONS
         TBOX_ASSERT(scalar_bconds.getSize() == NUM_3D_FACES);
#endif
         bscalarcase = scalar_bconds[bloc];
         refbdryloc = bloc;
      } else if (btype == EDGE3D_BDRY_TYPE) {
#ifdef DEBUG_CHECK_ASSERTIONS
         TBOX_ASSERT(scalar_bconds.getSize() == NUM_3D_EDGES);
#endif
         bscalarcase = scalar_bconds[bloc];
         refbdryloc = d_edge_bdry_face[bloc];
      } else { // btype == NODE3D_BDRY_TYPE
#ifdef DEBUG_CHECK_ASSERTIONS
         TBOX_ASSERT(scalar_bconds.getSize() == NUM_3D_NODES);
#endif
         bscalarcase = scalar_bconds[bloc];
         refbdryloc = d_node_bdry_face[bloc];
      }
#endif

      int num_bad_values = 0;

#if (NDIM == 2)
      num_bad_values =
      SkeletonBoundaryUtilities2::checkBdryData(
         d_uval->getName(),
         patch,
         vdb->mapVariableAndContextToIndex(d_uval, getDataContext()), 0,
         ghost_width_to_check,
         bbox,
         bscalarcase,
         d_bdry_edge_uval[refbdryloc]);
#endif
#if (NDIM == 3)
      num_bad_values =
      SkeletonBoundaryUtilities3::checkBdryData(
         d_uval->getName(),
         patch,
         vdb->mapVariableAndContextToIndex(d_uval, getDataContext()), 0,
         ghost_width_to_check,
         bbox,
         bscalarcase,
         d_bdry_face_uval[refbdryloc]);
#endif
#if (TESTING == 1)
      if (num_bad_values > 0) {
         tbox::perr << "\nMblkLinAdv Boundary Test FAILED: \n"
           << "     " << num_bad_values << " bad UVAL values found for\n"
           << "     boundary type " << btype << " at location " << bloc << endl;
      }
#endif

   }

}

hier::IntVector<NDIM> MblkLinAdv::getMultiblockRefineOpStencilWidth() const
{
   return (hier::IntVector<NDIM>(1));
}

hier::IntVector<NDIM> MblkLinAdv::getMultiblockCoarsenOpStencilWidth()
{
   return (hier::IntVector<NDIM>(0));
}
